import os
os.system('color')
from skyfield.api import load, Angle, Star, Topos, utc, wgs84
from skyfield import almanac
from skyfield.data import hipparcos
with load.open(hipparcos.URL) as f:
    df = hipparcos.load_dataframe(f)
import re, datetime as dt
from datetime import timezone, timedelta
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker
import matplotlib.dates as mdates

from matplotlib.patches import Ellipse
from tabulate import tabulate
import scipy.optimize as optimize
from utilities.utility_funcs import Utilities
from utilities.dead_reckoning import DRCalc

planets = load('de421.bsp')
ts = load.timescale()

class SightSession():
    """
    The Sight_session class takes all, of the pre-sight information that is relevant for the reduction process and
    passes it down to each Sight class. The idea is to represent an actual Sight-taking process. The Sight_session class
    gathers values that are only needed once, such as course or speed or index error.

    Parameters
    ----------
    date : np.array
        Covariance matrix from L-BFGS-B function
    time : tuple
        x, y position for center of ellipse
    dr_lat : int
        number of standard deviations.
    """
    dr_details = []
    num_of_sights = 0

    def __init__(self, data):
        # date, time,dr_lat,dr_long,course,speed,i_e,h_o_e,temp,pressure
        date, time, dr_lat, dr_long, course, speed, i_e, h_o_e, temp, pressure, fixdate, fixtime = data.split(',')

        self.date = date
        self.time = time
        self.course = Angle(degrees=(float(course)))
        self.speed = float(speed)
        deg, minutes, direction = dr_lat.split('-')
        self.dr_lat = (float(deg) + (float(minutes) / 60)) * (-1 if direction in 'S' else 1)
        deg, minutes, direction = dr_long.split('-')
        self.dr_long = (float(deg) + (float(minutes) / 60)) * (-1 if direction in 'W' else 1)
        self.i_e = Angle(degrees=(float(i_e) / 60))
        self.h_o_e = float(h_o_e)
        self.temp = float(temp)
        self.pressure = float(pressure)

        year, month, day = date.split('-')
        hour, minute, second = time.split(':')
        zd = int()
        tz = timezone(timedelta(hours=zd))
        self.datetime = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), tzinfo=tz)

        fix_year, fix_month, fix_day = fixdate.split('-')
        fixhr, fixmin, fixsec = fixtime.split(':')
        self.fixtime = dt.datetime(int(fix_year), int(fix_month), int(fix_day), int(fixhr), int(fixmin), int(fixsec),
                                   tzinfo=tz)
        SightSession.dr_details.append(
            [self.datetime, self.dr_lat, self.dr_long, self.course, self.speed, self.i_e, self.h_o_e, self.temp,
             self.pressure, self.fixtime])
        return

class Sight(SightSession):
    """
    The Sight class represents each individual sextant sight. It has a data/time a sextant Hs and a Body. Each Sight class
    instance is passed to the Sight_Reduction class.
    """
    num_of_sights = 0
    sight_times = []
    ho_array = []
    ho_vec_array = []
    sight_az_array = []
    gha_array_lop = []
    dec_array_lop = []
    body_array = []
    test_array_gha = []
    test_array_ho = []
    data_table = []

    computedlat = []
    computedlong = []
    hc_array = []
    intercept_array = []

    named_star_dict = {
        'Acamar': 13847, 'Achernar': 7588, 'Acrux': 60718, 'Adhara': 33579, 'Aldebaran': 21421, 'Algol': 14576,
        'Alioth': 62956,
        'Alkaid': 67301, 'Alnair': 109268, 'Alnilam': 26311, 'Alphard': 46390, 'Alphecca': 76267, 'Alpheratz': 677,
        'Altair': 97649,
        'Ankaa': 2081, 'Antares': 80763, 'Arcturus': 69673, 'Atria': 82273, 'Avior': 41037, 'Becrux': 62434,
        'Bellatrix': 25336,
        'Betelgeuse': 27989, 'Canopus': 30438, 'Capella': 24608, 'Deneb': 102098, 'Denebola': 57632, 'Diphda': 3419,
        'Dubhe': 54061,
        'Elnath': 25428, 'Enif': 107315, 'Eltanin': 87833, 'Fomalhaut': 113368, 'Gacrux': 61084, 'Gienah': 102488,
        'Hadar': 68702, 'Hamal': 9884, 'Kaus Australis': 90185, 'Kochab': 72607, 'Markab': 113963, 'Menkent': 68933,
        'Merak': 53910,
        'Miaplacidus': 45238, 'Mirach': 5447, 'Mirfak': 15863, 'Nunki': 92855, 'Peacock': 100751, 'Polaris': 11767,
        'Pollux': 37826,
        'Procyon': 37279, 'Rasalhague': 86032, 'Regulus': 49669, 'Rigel': 24436, 'Rigel Kent': 71683,
        'RigilKentaurus': 71683,
        'Sabik': 84012, 'Schedar': 3179, 'Shaula': 85927, 'Sirius': 32349, 'Spica': 65474, 'Suhail': 44816,
        'Vega': 91262,
        'Zubenelgenubi': 72622,
    }

    def __init__(self, data):

        body, hs, date, time = data.split(',')
        self.body = body
        self.date = date
        self.time = time
        hs_deg, hs_min = hs.split('-')
        hs = (float(hs_deg) + (float(hs_min) / 60))
        self.hs = Angle(degrees=(hs))
        year, month, day = date.split('-')
        hour, minute, second = time.split(':')
        self.datetime = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), tzinfo=utc)
        self.t = ts.utc(self.datetime)
        Sight.num_of_sights += 1
        # plot.session.num_of_sights += 1

        self.get_dr_time_delta()
        self.get_dr_details()
        self.get_sight_dr_positions()

        self.compute_gha_dec()
        self.dip_correction()
        self.index_correction()
        self.ha_calc()
        self.get_HP()
        self.parallax_altitude_correction()
        self.semi_diameter_correction()
        self.refraction()
        self.ho_constructor()
        self.hc_constructor()
        self.intercept()

        self.sight_data = [f'{Sight.num_of_sights - 1}',
                           f'{self.body.upper()}',
                           Utilities.print_position(self.computed_lat, latitude=True),
                           Utilities.print_position(self.computed_long, latitude=False),
                           f'{self.time}',
                           f'{Utilities.hmt_str(self.GHA.degrees)}',
                           f'{Utilities.hmt_str(self.DEC.degrees)}',
                           f'{round(self.AZ.degrees, )}',
                           f'{Utilities.hmt_str(self.ho.degrees)}',
                           f'{Utilities.hmt_str(self.hc.degrees)}',
                           f'{format(self.int, ".1f")}'
                           ]

        Sight.data_table.append(self.sight_data)
        Sight.sight_times.append(self.datetime)
        Sight.body_array.append(self.body)

        self.array_creation()

        Sight.ho_array.append(self.ho_array)
        Sight.sight_az_array.append(self.AZ)

        return

    def get_dr_time_delta(self):
        """Computes the time delta in seconds between the DR time and the time of the Sight"""
        self.dr_time = SightSession.dr_details[0][0]
        self.sighttime = self.datetime
        self.drtimedelta = dt.timedelta.total_seconds(self.sighttime - self.dr_time)

        return

    def get_dr_details(self):
        """Fetches DR Lat, Long, Course, Speed from Sight_session class dr_details list"""
        self.dr_lat = SightSession.dr_details[0][1]
        self.dr_long = SightSession.dr_details[0][2]
        self.course = SightSession.dr_details[0][3]
        self.speed = SightSession.dr_details[0][4]

        return

    def get_sight_dr_positions(self):
        """Uses get_dr_time_delta and get_dr_details functions to provide information to dr_calc class"""
        lat = self.dr_lat
        long = self.dr_long
        course = self.course
        speed = self.speed
        timed = self.drtimedelta
        self.computed_lat = DRCalc(lat, long, timed, course.degrees, speed).drlatfwds
        self.computed_long = DRCalc(lat, long, timed, course.degrees, speed).drlongfwds
        Sight.computedlat.append(self.computed_lat)
        Sight.computedlong.append(self.computed_long)

        return

    def compute_gha_dec(self):
        """Functionally identical to Get_GHA_DEC function in utilities, just called every time a Sight Object
         is initialized"""
        body = self.body
        if body == 'SunLL' or body == 'SunUL':
            celestial_body = planets['Sun']
        elif body == 'MoonLL' or body == 'MoonUL':
            celestial_body = planets['Moon']
        elif body == 'Mars':
            celestial_body = planets['Mars']
        elif body == 'Venus':
            celestial_body = planets['Venus']
        elif body == 'Jupiter':
            celestial_body = planets['Jupiter Barycenter']
        elif body == 'Saturn':
            celestial_body = planets['Saturn Barycenter']
        elif body == 'Uranus':
            celestial_body = planets['Uranus Barycenter']
        elif body == 'Mercury':
            celestial_body = planets['Mercury']
        else:
            which_star = body
            hid = Sight.named_star_dict.get(which_star)
            celestial_body = Star.from_dataframe(df.loc[hid])
        obs = planets['Earth']
        # datetime object of sight
        dt = self.t.utc_datetime()
        # time delta between DR dateime object and Sight datetimeobject

        position = obs + Topos(latitude_degrees=(self.computed_lat), longitude_degrees=(self.computed_long))

        astro = position.at(self.t).observe(celestial_body)
        app = astro.apparent()

        astrometric = obs.at(self.t).observe(celestial_body)
        apparent = obs.at(self.t).observe(celestial_body).apparent()
        alt, az, distance = app.altaz()
        ra, dec, distance, = apparent.radec(epoch='date')

        ghaa = Angle(degrees=(self.t.gast) * 15)
        gha = Angle(degrees=((self.t.gast - ra.hours) * 15 % 360 - 0))
        self.GHA = gha
        self.DEC = dec
        self.ALT = alt
        self.AZ = az
        self.DIST = distance
        self.ghaa = ghaa

        return

    def dip_correction(self):
        """Uses height of eye in feet provided in Sight_session.dr_details to compute dip correction"""
        dip_corr = Angle(degrees=(-1 * (.97 * np.sqrt(SightSession.dr_details[0][6])) / 60))
        self.dip_corr = dip_corr

        return

    def index_correction(self):
        """Uses user provided index error in arc minutes provided in Sight_session.dr_details to compute dip
        correction """
        index_corr = SightSession.dr_details[0][5]
        self.index_corr = index_corr

        return

    def ha_calc(self):
        """Calculates Ha in degrees from hs, dip and index correction functions"""
        ha = Angle(degrees=(self.hs.degrees + self.dip_corr.degrees + self.index_corr.degrees))
        self.ha = ha

        return

    def get_HP(self):
        """Calculates Horizontal Parallax in degrees for Sun, Moon. If Venus, Saturn, Mars or Jupiter are provided,
        it uses the same formula to calculate HP as the difference is not noticeable for celestial navigation"""
        body = self.body
        if body == 'SunLL' or body == 'SunUL':
            self.hp_degrees = Angle(degrees=(0.0024))
        elif body == 'MoonLL' or body == 'MoonUL':
            distance_rad = np.deg2rad(self.DIST.km)
            hp_numerator = np.deg2rad(6378.14)
            hp_rad = np.arcsin(hp_numerator / distance_rad)
            hp_degrees = np.rad2deg(hp_rad)
            self.hp_degrees = Angle(degrees=hp_degrees)
        elif body == 'Venus' or body == 'Saturn' or body == 'Mars' or body == 'Jupiter':
            distance = self.DIST.km
            self.hp_degrees = Angle(degrees=(1.315385814 * 10 ** 9 / distance) / 3600)

        else:
            self.hp_degrees = Angle(degrees=(0))
        return

    def parallax_altitude_correction(self):
        """Calculates the parallax altitude correction in degrees using the get_HP function"""
        body = self.body

        if body == 'Venus' or body == 'Mars':
            parallax_corr = Angle(degrees=(self.hp_degrees.degrees * np.cos(self.ha.radians) * (
                    1 - (np.sin(np.deg2rad(SightSession.dr_details[0][1])) ** 2.0) / 297.0)))
            self.parallax_corr = parallax_corr

        elif body == 'SunLL' or body == 'SunUL':
            parallax_corr = Angle(degrees=(self.hp_degrees.degrees * np.cos(self.ha.radians) *
                                           (1 - (np.sin(np.deg2rad(SightSession.dr_details[0][1])) ** 2.0) / 297.0)))
            self.parallax_corr = parallax_corr

        elif body == 'MoonLL' or body == 'MoonUL':

            OB = Angle(degrees=(-0.0017 * np.cos(self.ha.radians)))
            parallax_corr = Angle(degrees=(self.hp_degrees.degrees * np.cos(self.ha.radians) *
                                           (1 - (np.sin(np.deg2rad(SightSession.dr_details[0][1])) ** 2.0) / 297.0)))

            self.parallax_corr = Angle(degrees=parallax_corr.degrees + OB.degrees)

        else:
            parallax_corr = Angle(degrees=(0))
            self.parallax_corr = parallax_corr

        return

    def semi_diameter_correction(self):
        """Calculates semi-diameter correction in degrees for Sun or Moon"""
        body = self.body

        if body == 'SunLL' or body == 'SunUL':
            sds = Angle(degrees=((15.9938 / self.DIST.au) / 60))
            if body == 'SunLL':
                self.sd_corr = Angle(degrees=(sds.degrees))
            else:
                self.sd_corr = Angle(degrees=(-1 * sds.degrees))

        elif body == 'MoonLL' or body == 'MoonUL':
            sdm = Angle(degrees=(.272476 * (self.hp_degrees.degrees)))
            sdm_tc = Angle(degrees=(sdm.degrees * (1 + np.sin(self.ALT.radians) / 60.27)))
            if body == 'MoonLL':
                self.sd_corr = Angle(degrees=(sdm_tc.degrees))
            else:
                self.sd_corr = Angle(degrees=(-1 * sdm_tc.degrees))
        else:
            self.sd_corr = Angle(degrees=(0))

        return

    def refraction(self):
        """Calculates refraction correction in degrees for celestial object"""
        pmb = SightSession.dr_details[0][8]
        TdegC = SightSession.dr_details[0][7]
        f = 0.28 * pmb / (TdegC + 273.0)
        ro = -1 * ((0.97127 / np.tan(self.ha.radians)) - (0.00137 / (np.tan(self.ha.radians)) ** 3)) / 60
        self.ref = Angle(degrees=(ro * f))

        return

    def ho_constructor(self):
        """Calculates Ho in degrees using the ha_calc, refraction, semi_diameter_correction and
        parallax_altitude_correction functions """
        ho = Angle(degrees=(self.ha.degrees + self.ref.degrees + self.sd_corr.degrees + self.parallax_corr.degrees))
        self.ho = ho
        return

    def hc_constructor(self):
        """Calculates Hc in degrees using lat and long provided by the get_sight_dr_positions function """
        lha = Angle(degrees=(self.computed_long + self.GHA.degrees) % 360)
        lat = Angle(degrees=self.computed_lat)

        self.hc = Angle(radians=(np.arcsin(
            np.sin(lat.radians) * np.sin(self.DEC.radians) + np.cos(lat.radians) * np.cos(self.DEC.radians) * np.cos(
                lha.radians))))

        return

    def intercept(self):
        """Calculates the Marc St.Hillaire intercept in minutes using the ho_constructor and hc_constructor methods.
        This is for the navigators reference only and isn't used by the internal position calculation. """
        intercept = (self.ho.degrees) - self.hc.degrees
        self.int = float(intercept * 60)
        Sight.intercept_array.append(intercept)

        return

    def array_creation(self):
        """appends to Sight.ho_array, Sight.gha_array_lop and Sight.dec_array_lop"""
        ho_array = np.array([(self.ho.degrees)])
        self.ho_array = ho_array
        Sight.gha_array_lop.append(self.GHA.radians)
        Sight.dec_array_lop.append(self.DEC.radians)

        return


class SightReduction(Sight):
    """ The main Sight Reduction algorithm and plotting algorithms live here. SightReduction() doesn't fire unless
    it is instantiated as True, and will not work unless the required arrays in Sight() and SightSession() are
    filled. For multiple Sight Reductions (essential for the iterative recomputations), the arrays need to be reset
    to empty every time, this happens in main.py currently.

    Potential Improvements:
    1. Remove plotting functionality entirely
    2. Reset arrays internal to cnav.py
    """
    time_delta_array = []
    ho_corrections_array = []
    final_ho_array = []
    position_array_l = []
    position_array_lon = []
    latx_lists = []
    longx_lists = []
    ho_array_rfix = []
    pos_array_lop_lon = []
    pos_array_lop_lat = []
    final_position_array = []
    sight_anl_table = []
    gui_position_table = []

    def __init__(self, reduction):
        self.reduction = reduction
        self.last_time = None
        self.last_time_sort()
        self.ho_correction()
        self.final_ho_sr()
        self.vector_reduction()
        self.sight_analysis()
        # self.error_trapping()

        self.bx_method()
        self.scatter_plot_analyzer()
        self.lop_plot()

    def last_time_sort(self):
        """Computes the time delta in total seconds between the user provided time of fix and the time of each sight
        object, appends to Sight_Reduction.time_delta_array
        """
        fix_time = SightSession.dr_details[0][9]
        for i in range(Sight.num_of_sights):
            SightReduction.time_delta_array.append(dt.timedelta.total_seconds(fix_time - Sight.sight_times[i]))
        return

    def ho_correction(self):
        """advances/retards the sight lop by changing the Ho value using DR course/speed information.

        Parameters
        ----------
        Sight_session.dr_details[0][4] : float
            DR Speed information from Sight_session object
        Sight_session.dr_details[0][3]: float
            DR Course information in radians
        Sight.sight_az_array : float
            Angle object computed in Sight.Get_GHADEC

        """
        for i in range(Sight.num_of_sights):
            SightReduction.ho_corrections_array.append((SightSession.dr_details[0][4]
                                                        * (SightReduction.time_delta_array[i] / 3600)) / 60
                                                       * np.cos(Sight.sight_az_array[i].radians
                                                                 - SightSession.dr_details[0][3].radians))

        return

    def final_ho_sr(self):
        """Sums Sight.ho_array and Sight_Reduction.ho_corrections_array to create Sight_Reduction.ho_array_rfix,
        an array of Ho's corrected for the movement of the vessel to compute a running fix. """
        for i in range(Sight.num_of_sights):
            SightReduction.ho_array_rfix.append(
                np.deg2rad(Sight.ho_array[i] + SightReduction.ho_corrections_array[i]))
        return

    latitude_array = []
    longitude_array = []
    test_array = []

    def obj_function(self, params):
        """Objective Function to be minimized."""
        int_sum = []
        lat, long = params
        for i in range(len(Sight.body_array)):
            gha = Sight.gha_array_lop[i]
            dec = Sight.dec_array_lop[i]
            lha = (long + np.rad2deg(gha)) % 360
            hc = np.arcsin(np.sin(np.deg2rad(lat)) * np.sin(dec) + 
                           np.cos(np.deg2rad(lat)) * np.cos(dec) * np.cos(np.deg2rad(lha)))
            ho = SightReduction.ho_array_rfix[i]
            intercept = ho - hc
            int_sum.append(intercept ** 2)
        return np.sqrt(np.sum(int_sum) / Sight.num_of_sights)

    def calculate_initial_guess(self):
        """Calculate initial guess for optimization."""
        dr_details = SightSession.dr_details[0]
        dr_calc_lat = DRCalc(dr_details[1], dr_details[2],
                             timedelta.total_seconds(dr_details[9] - dr_details[0]),
                             dr_details[3].degrees, dr_details[4])
        dr_lat = dr_calc_lat.drlatfwds  # Assuming drlatfwds returns forward latitude
        dr_long = dr_calc_lat.drlongfwds  # Assuming drlongfwds returns forward longitude
        return dr_lat, dr_long

    def vector_reduction(self):
        """Computes fix using optimization given the Sight and SightSession information."""
        self.dr_lat, self.dr_long = self.calculate_initial_guess()

        # Use genetic algorithm for optimization
        self.res = optimize.differential_evolution(self.obj_function, bounds=[(0, 90), (-180, 180)])

        # Process optimization results
        self.process_optimization_results()

        # Update computed values and format for display
        self.update_computed_values()

        # Append results to class arrays
        self.append_results()

    def process_optimization_results(self):
        """Process the results from the optimization."""
        # Create table for scipy optimization results
        headers = ["Success", "Iterations", "Func. Value"]
        res_info = [[self.res.success, self.res.nit, self.res.fun]]
        self.res_info_str = tabulate(res_info, headers=headers)

        # Calculate errors from inverse Hessian
        try:
            hess_inv_diag = np.sqrt(np.diag(self.res.hess_inv.todense()))
            self.latitude_error, self.longitude_error = hess_inv_diag
        except Exception as e:
            print(f"Error processing Hessian: {e}")
            self.latitude_error = self.longitude_error = None

    def update_computed_values(self):
        """Update computed values for latitude and longitude."""
        self.fit_latitude = Angle(degrees=(self.res.x[0]))
        self.fit_longitude = Angle(degrees=(self.normalize_longitude(self.res.x[1])))

    def normalize_longitude(self, longitude):
        """Normalize longitude to be within -180 to 180 degrees."""
        if longitude > 180:
            return longitude - 360
        elif longitude < -180:
            return longitude + 360
        return longitude

    def append_results(self):
        """Append results to class arrays."""
        lat_sign = 'N' if self.fit_latitude.degrees >= 0 else 'S'
        long_sign = 'E' if self.fit_longitude.degrees >= 0 else 'W'

        self.print_latitude = self.fit_latitude if lat_sign == 'N' else Angle(degrees=abs(self.fit_latitude.degrees))
        self.print_longitude = self.fit_longitude if long_sign == 'E' else Angle(degrees=abs(self.fit_longitude.degrees))

        self.final_l_string = f'{Utilities.hmt_str(self.print_latitude.degrees)} {lat_sign}'
        self.final_lon_string = f'{Utilities.hmt_str(self.print_longitude.degrees)} {long_sign}'

        SightReduction.position_array_l.append(self.final_l_string)
        SightReduction.position_array_lon.append(self.final_lon_string)
        SightReduction.pos_array_lop_lat.append(self.fit_latitude.degrees)
        SightReduction.pos_array_lop_lon.append(self.fit_longitude.degrees)

        self.fixtime = SightSession.dr_details[0][9]

    sight_analysis_lat_time_of_sight = []
    sight_analysis_long_time_of_sight = []
    sight_analysis_lat_plus_one = []
    sight_analysis_long_plus_one = []
    sight_analysis_lat_minus_one = []
    sight_analysis_long_minus_one = []
    hc_timeofsight = []
    hc_plusone = []
    hc_minusone = []
    stats_table_2 = []

    ####

    drsight_analysis_lat_time_of_sight = []
    drsight_analysis_long_time_of_sight = []
    drsight_analysis_lat_plus_one = []
    drsight_analysis_long_plus_one = []
    drsight_analysis_lat_minus_one = []
    drsight_analysis_long_minus_one = []
    drhc_timeofsight = []
    drhc_plusone = []
    drhc_minusone = []

    def calculate_position(self, lat, long, time_delta, course, speed):
        dr_calc = DRCalc(lat, long, time_delta, course, speed)
        return dr_calc.drlatbackwards, dr_calc.drlongbackwards

    def calculate_heavenly_body_info(self, body, datetime, lat, long):
        ephem = Utilities.get_gha_dec(body, datetime, lat, long)
        gha, dec = ephem[0], ephem[1]
        lat_hc = Angle(degrees=lat)
        long_hc = Angle(degrees=long)
        lha = Angle(degrees=(gha.degrees + long_hc.degrees))
        hc = np.arcsin((np.sin(lat_hc.radians) * np.sin(dec.radians)) + 
                       (np.cos(lat_hc.radians) * np.cos(dec.radians) * np.cos(lha.radians)))
        return np.rad2deg(hc)

    def sight_analysis(self):
        lat = float(self.fit_latitude.degrees)
        long = float(self.fit_longitude.degrees)
        course = SightSession.dr_details[0][3].degrees
        speed = SightSession.dr_details[0][4]

        for i in range(Sight.num_of_sights):
            time_delta = self.time_delta_array[i]
            sight_time = Sight.sight_times[i]
            body = Sight.body_array[i]

            # Calculate positions for exact, +1 minute, and -1 minute times
            lat_time_of_sight, long_time_of_sight = self.calculate_position(lat, long, time_delta, course, speed)
            lat_plus_one, long_plus_one = self.calculate_position(lat, long, time_delta + 60, course, speed)
            lat_minus_one, long_minus_one = self.calculate_position(lat, long, time_delta - 60, course, speed)

            # Store calculated positions
            self.sight_analysis_lat_time_of_sight.append(lat_time_of_sight)
            self.sight_analysis_long_time_of_sight.append(long_time_of_sight)
            self.sight_analysis_lat_plus_one.append(lat_plus_one)
            self.sight_analysis_long_plus_one.append(long_plus_one)
            self.sight_analysis_lat_minus_one.append(lat_minus_one)
            self.sight_analysis_long_minus_one.append(long_minus_one)

            # Calculate and store heavenly body information
            self.datetime = sight_time
            self.hc_timeofsight.append(self.calculate_heavenly_body_info(body, self.datetime, lat_time_of_sight, long_time_of_sight))
            self.datetime = sight_time + dt.timedelta(seconds=60)
            self.hc_plusone.append(self.calculate_heavenly_body_info(body, self.datetime, lat_plus_one, long_plus_one))
            self.datetime = sight_time - dt.timedelta(seconds=60)
            self.hc_minusone.append(self.calculate_heavenly_body_info(body, self.datetime, lat_minus_one, long_minus_one))

    d_array = []

    
    def y_fmt(self, y, x):
        """Format matplotlib y-value ("dd°mm')"""
        return Utilities.hmt_str(y)

    def datetime_to_float(self, d):
        """Convert datetime to timestamp"""
        return d.timestamp()

    def setup_subplot(self, num_of_sights, index):
        """Setup a subplot based on the number of sights and current index"""
        if num_of_sights % 2 == 0:
            gs = gridspec.GridSpec(2, num_of_sights // 2)
        elif num_of_sights < 6:
            gs = gridspec.GridSpec(2, 3)
        elif 6 < num_of_sights < 9:
            gs = gridspec.GridSpec(2, 4)
        else:
            gs = gridspec.GridSpec(4, 4)

        ax = plt.subplot(gs[index])
        ax.yaxis.set_major_formatter(mticker.FuncFormatter(self.y_fmt))
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%M:%S'))
        ax.set_facecolor('#212946')
        ax.grid(color='#2A3459')
        ax.tick_params(axis='x', rotation=-45, labelsize=6)
        ax.tick_params(axis='y', rotation=45, labelsize=6)
        return ax

    def scatter_plot_analyzer(self):
        """Matplotlib plotting function for sight_analysis."""
        d_dict = {}
        top_unique_indexes = []
        plt.style.use('dark_background')
        plt.subplots(figsize=[7, 7])
        plt.figure(1)
        plt.clf()

        for i in range(Sight.num_of_sights):
            one = SightReduction.hc_plusone[i]
            two = SightReduction.hc_minusone[i]
            three = Sight.ho_array[i]
            four = SightReduction.hc_timeofsight[i]

            time1 = Sight.sight_times[i]
            time_before = time1 + dt.timedelta(seconds=60)
            time_after = time1 - dt.timedelta(seconds=60)

            x = np.array([time_before, time1, time_after])
            y = np.array([two, four, one])

            ax = self.setup_subplot(Sight.num_of_sights, i)

            ax.plot(x, y)
            ax.scatter(time1, three, color='red')

            p1 = np.array([self.datetime_to_float(time_after), one], dtype=object)
            p2 = np.array([self.datetime_to_float(time_before), two], dtype=object)
            p3 = np.array([self.datetime_to_float(time1), three], dtype=object)
            p4 = np.array([self.datetime_to_float(time1), four], dtype=object)

            d = float((np.cross(p2 - p1, p3 - p1) / np.linalg.norm(p2 - p1)) * 60)
            SightReduction.d_array.append(d)

            ax.set_title(f"{Sight.body_array[i]} || # {i + 1} || Scatter: %.2f' " % d, size=8, color='#f39c12')
            ax.text(time1.minute + .1, three, f'{Utilities.hmt_str(three)}', )
            ax.tick_params(axis='both', which='major', labelsize=6)

            d_dict[i] = d

        # Adjust subplot spacing based on the number of sights
        if Sight.num_of_sights % 2 == 0:
            plt.subplots_adjust(left=.062, bottom=.062, right=.97, top=.917, wspace=.2, hspace=.2)
        elif Sight.num_of_sights < 6:
            plt.subplots_adjust(left=.062, bottom=.062, right=.97, top=.917, wspace=.2, hspace=.2)
        else:
            plt.subplots_adjust(left=.057, bottom=.052, right=.979, top=.93, wspace=.248, hspace=.42)
            
       
        # Sorts d values closest to 0
        sorted_values = sorted(d_dict.values(), key=lambda x: abs(x))           
        sorted_dict = {}
        for i in sorted_values:
            for k in d_dict.keys():
                if d_dict[k] == i:
                    sorted_dict[k] = d_dict[k]
                    break
        pairs_recieved = []

        for i in sorted_dict.keys():
            pair = [Sight.body_array[i], i, Sight.sight_times[i]]
            pairs_recieved.append(pair)

        # just the body name
        unique_list = []
        unique_timechunks = []

        # traverse for all elements
        for x in Sight.body_array:
            # check if exists in unique_list or not
            if x not in unique_list:
                unique_list.append(x)

        # print list
        pairs_recieved.sort(key=lambda x: x[0])
        singlebodyelementarray = []
        if len(unique_list) == 1 and Sight.num_of_sights <= 3:
            self.top_unique = pairs_recieved

        elif len(unique_list) == 1 and Sight.num_of_sights > 3:
            try:
                for i in d_dict.keys():
                    singlebodyelement = (Sight.body_array[i], i, SightReduction.d_array[i], Sight.sight_times[i])
                    singlebodyelementarray.append(singlebodyelement)

                singlebody = Utilities.single_body_time_divide(singlebodyelementarray)

                self.top_unique = singlebody

            except:
                pass

        else:
            self.top_unique = []

            for x in pairs_recieved:
                if x[0] in unique_list:
                    self.top_unique.append(x)
                    top_unique_indexes.append(x[1])
                    unique_list.remove(x[0])

        top_unique_indexes.sort()

        return

    def calculate_lop_coordinates(self, dec, gha, ho, Bx_r, long):
        """Calculate latitude and longitude coordinates for a point on the LOP."""
        Lx_r = np.arcsin(np.sin(dec) * np.sin(ho) + np.cos(dec) * np.cos(ho) * np.cos(Bx_r))
        Lx_d = np.rad2deg(Lx_r)

        LHAx_r = np.arcsin((np.cos(ho) * np.sin(Bx_r)) / np.cos(Lx_r))
        LHAx_d = np.rad2deg(LHAx_r) % 360

        Longx_d = (LHAx_d + np.rad2deg(gha)) % 360
        Longx_d_2 = (np.rad2deg(gha) - LHAx_d)

        longitude_buffet = [360 - Longx_d, abs(Longx_d_2 + 180), (abs(Longx_d_2 + 180)) * -1, 
                            Longx_d * -1, Longx_d * -1, 180 - Longx_d_2, Longx_d_2]
        longitude_buffet = sorted(longitude_buffet)

        closest_long = min(longitude_buffet, key=lambda x: abs(x - long))

        return Lx_d, closest_long

    def bx_method(self):
        """Plots LOPs by computing pairs of x,y coordinates that lie on the LOP."""
        for i in range(Sight.num_of_sights):
            dec = Sight.dec_array_lop[i]
            gha = Sight.gha_array_lop[i]
            long = SightReduction.pos_array_lop_lon[0]
            lat = np.deg2rad(SightReduction.pos_array_lop_lat[0])
            lha = np.deg2rad(((np.rad2deg(gha) + long) % 360))
            hc_rad = np.arcsin((np.sin(dec) * np.sin(lat)) + (np.cos(dec) * np.cos(lat) * np.cos(lha)))

            z_rad = np.arccos((np.sin(lat) - np.sin(dec) * np.sin(hc_rad)) / (np.cos(dec) * np.cos(hc_rad)))
            z_rad = (360 - np.rad2deg(z_rad)) if np.rad2deg(lha) < 180 else np.rad2deg(z_rad)

            Bx_r_i = np.deg2rad(z_rad)

            latx_list = []
            longx_list = []
            ho = SightReduction.ho_array_rfix[i]

            # Calculate coordinates for both sides of Bx
            for Bx_r in [Bx_r_i + np.deg2rad(.75), Bx_r_i - np.deg2rad(.75)]:
                Lx_d, closest_long = self.calculate_lop_coordinates(dec, gha, ho, Bx_r, long)
                latx_list.append(Lx_d)
                longx_list.append(closest_long)

            SightReduction.latx_lists.append(latx_list)
            SightReduction.longx_lists.append(longx_list)

        return        
    
    def lop_plot(self):

        plt.style.use('dark_background')
        # plt.subplots(figsize=[6, 6])
        plt.figure(2)
        plt.clf()

        self.ax = plt.subplot(111)
        plt.subplots_adjust(left=.148, bottom=.121, right=.957, top=.929, wspace=.2, hspace=.2)

        self.ax.set_facecolor('#212946')

        # very light grey
        self.ax.grid(color='#2A3459')

        # latitude
        def y_fmt(x, y):
            """Uses Utilities.print_position to format matplotlib y-value ("dd°mm' N/S)
            Parameters
            ----------
            x : float
            matplotlib latitude value in dd.dddd
            """
            return Utilities.print_position(x, latitude=True)

        # longitude
        def x_fmt(y, x):
            """Uses Utilities.print_position to format matplotlib y-value ("ddd°mm' E/W) and constrains it to < 180 deg.
            Parameters
            ----------
            y : float
            matplotlib longitude value in ddd.dddd
            """
            if y > 180:
                y = y - 360
            elif y < -180:
                y = y + 360
            return Utilities.print_position(y, latitude=False)

        self.ax.yaxis.set_major_formatter(mticker.FuncFormatter(y_fmt))
        self.ax.xaxis.set_major_formatter(mticker.FuncFormatter(x_fmt))
        plt.xticks(rotation=45)

        for i in range(len(SightReduction.latx_lists)):
            y = SightReduction.latx_lists[i]
            x = SightReduction.longx_lists[i]

            plt.plot(SightReduction.longx_lists[i], SightReduction.latx_lists[i],
                     label=f'{Sight.body_array[i]} {Sight.sight_times[i]}')
            plt.legend(prop={'size': 7})
            plt.text(x[0] + .05, y[0] + .01, f'{Sight.body_array[i]}', size=9, color='#fff6')

        plt.scatter(SightReduction.pos_array_lop_lon[0], SightReduction.pos_array_lop_lat[0], marker='o', color='red')

        try:
            self.err_ellipse = Utilities.plot_cov_ellipse(self.res.hess_inv.todense(), (self.res.x[1], self.res.x[0]),
                                                        ax=self.ax, fc='none', edgecolor='#f39c12')

            self.ax.add_patch(self.err_ellipse)
        except:
            pass
 
        plt.scatter(self.dr_long, self.dr_lat, marker='+')
        plt.text(self.dr_long, self.dr_lat, f'{SightSession.dr_details[0][9].strftime("%H:%M")} UTC DR', size=9,
                 color='#00bc8c')

        plt.xlabel('Longitude', size=8, color='#fff6')
        plt.ylabel('Latitude', size=8, color='#fff6')

        plt.title(f'Computed Fix: {self.final_l_string} {self.final_lon_string}')

        final_position = [self.final_l_string, self.final_lon_string]
        SightReduction.final_position_array.append(final_position)

        # stats_header2 = ["N/S ERROR 68% PROB.", "E/W ERROR 68% PROB.", "N/S ERROR 95% PROB.", "E/W ERROR 95% PROB.",
        #                  "Sys. Err."]

        # stats_table_2 = [[f"(+/-) {np.round(((self.latitude_error / 2)), 2)}",
        #                   f" (+/-) {np.round(((self.longitude_error * np.cos(np.deg2rad(self.res.x[0])) / 2)), 2)}",
        #                   f"(+/-) {np.round(((self.latitude_error)), 2)}",
        #                   f" (+/-) {np.round(((self.longitude_error * np.cos(np.deg2rad(self.res.x[0])))), 2)}",
        #                   f'{np.round(np.mean(SightReduction.d_array), 2)}']]

        # SightReduction.stats_table_2.append(stats_table_2)

        # print(tabulate(stats_table_2, stats_header2, tablefmt='github'))
         
        anl_table = []
        anl_headers = ["BODY", "INDEX", "TIME"]
        for x in self.top_unique:
            string = [x[0], x[1],
                      Sight.sight_times[x[1]].strftime("%Y-%m-%d %H:%M:%S UTC")]
            anl_table.append(string)

        SightReduction.sight_anl_table.append(anl_table)
        sight_anl_tbl = tabulate(anl_table, anl_headers, tablefmt='github')
        # print(sight_anl_tbl)

        headers = ["INDEX", "BODY", "DR L", "DR λ", "TIME", 'GHA', 'DEC', 'AZ', 'Ho', 'Hc', 'Int.']

        self.bigdata = (tabulate(Sight.data_table, headers, tablefmt='rst'))
        gui_position_data = (
            SightSession.dr_details[0][9].strftime("%Y-%m-%d %H:%M:%S"), self.final_l_string, self.final_lon_string,
            Utilities.print_position(self.dr_lat, latitude=True), Utilities.print_position(self.dr_long, latitude=False))
        SightReduction.gui_position_table.append(gui_position_data)

        # print(self.bigdata)

        position_heads = ['Date', 'Computed Lat', 'Computed Long', 'DR Lat', 'DR Long']
        # print(tabulate([gui_position_data], position_heads, tablefmt='github'))
        ####################################################################################


import os
import tkinter as tk
import ttkbootstrap as ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import utilities.celestial_engine as celestial_engine
from utilities.sight_planning import SightSessionPlanning
from gui_pages.sight_entry_page import SightEntryPage
from utilities.autocompletion import AutoComplete
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.tooltip import ToolTip
from utilities.tooltips import TextExtractor

class SightPlanningPage(ttk.Frame):
    def __init__(self, container, PageOne):
        super().__init__(container)
        self.gui_page1 = PageOne

        self.create_label_frames()
        self.create_notebook()
        self.create_planning_controls_page()
        self.create_time_of_phenomena_treeview()
        self.create_planning_treeview()
        self.create_optimal_triad_treeview()
        self.autocompletion_binding()
        self.create_tooltips()

         # Configure the weights for the SightPlanningPage frame
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

      
        
    def create_label_frames(self):
        self.sight_planning_frame = ttk.LabelFrame(self, text='Sight Planning')

        # grid, fill both directions
        self.sight_planning_frame.grid(row=0, column=0, sticky='NESW')

        # Configure the weights
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)


    def create_notebook(self):
        self.notebook = ttk.Notebook(self.sight_planning_frame)
        
        # grid
        self.notebook.grid(row=0, column=0, sticky='NESW')

        # Configure the weights
        self.sight_planning_frame.grid_rowconfigure(0, weight=1)
        self.sight_planning_frame.grid_columnconfigure(0, weight=1)


        self.page1 = ttk.Frame(self.notebook)
        self.page1.grid(sticky='nsew')  # Use grid instead of pack
        self.page1.grid_rowconfigure(0, weight=1)
        self.page1.grid_columnconfigure(0, weight=1)

        self.page2 = ttk.Frame(self.notebook)
        self.page2.grid(sticky='nsew')  # Use grid instead of pack
        self.page2.grid_rowconfigure(0, weight=1)
        self.page2.grid_columnconfigure(0, weight=1)

        self.page3 = ttk.Frame(self.notebook)
        self.page3.grid(sticky='nsew')  # Use grid instead of pack
        self.page3.grid_rowconfigure(0, weight=1)
        self.page3.grid_columnconfigure(0, weight=1)

        self.page4 = ttk.Frame(self.notebook)
        self.page4.grid(sticky='nsew')  # Use grid instead of pack
        self.page4.grid_rowconfigure(0, weight=1)
        self.page4.grid_columnconfigure(0, weight=1)

        self.notebook.add(self.page1, text='Planning Control')
        self.notebook.add(self.page3, text='Visible Bodies')
        self.notebook.add(self.page4, text='Optimal Triads')

        self.page1.master = self
        self.page2.master = self
        self.page3.master = self
        self.page4.master = self


    def create_planning_controls_page(self):
        # create label frames
        self.planning_controls_frame = ttk.LabelFrame(self.page1, text='Planning Controls')
        
        # grid
        self.planning_controls_frame.grid(row=0, column=0)

        self.phenoma_frame = ttk.LabelFrame(self.page1, text='Phenomena')
        
        # grid
        self.phenoma_frame.grid(row=1, column=0, sticky='NESW')

        # create label and entry fields for date, time, dr latitude, dr longitude
        # Date
        self.date_label = ttk.Label(self.planning_controls_frame, text='Date')
        self.date_label.grid(row=0, column=0, padx=10, pady=10, sticky='W')
        self.date_entry = ttk.Entry(self.planning_controls_frame, width = 12)
        self.date_entry.grid(row=0, column=1, padx=10, pady=10, sticky='W')

  
        # Time
        self.time_label = ttk.Label(self.planning_controls_frame, text='Time')
        self.time_label.grid(row=1, column=0, padx=10, pady=10, sticky='W')
        self.time_entry = ttk.Entry(self.planning_controls_frame, width = 12)
        self.time_entry.grid(row=1, column=1, padx=10, pady=10, sticky='W')

  
        # labels and Inputs for DR Latitude, Longitude
  
        # DR Latitude
        self.dr_latitude_label = ttk.Label(self.planning_controls_frame, text='DR Latitude')
        self.dr_latitude_label.grid(row=2, column=0, padx=10, pady=10, sticky='W')

        self.dr_latitude_entry = ttk.Entry(self.planning_controls_frame, width = 12)
        self.dr_latitude_entry.grid(row=2, column=1, padx=10, pady=10, sticky='W')

  
        # DR Longitude
        self.dr_longitude_label = ttk.Label(self.planning_controls_frame, text='DR Longitude')
        self.dr_longitude_label.grid(row=3, column=0, padx=10, pady=10, sticky='W')

  
        self.dr_longitude_entry = ttk.Entry(self.planning_controls_frame, width = 12)
        self.dr_longitude_entry.grid(row=3, column=1, padx=10, pady=10, sticky='W')

        # make entry fields helvetica, 12, bold
        self.dr_latitude_entry.configure(font=('Helvetica', 12, 'bold'))
        self.dr_longitude_entry.configure(font=('Helvetica', 12, 'bold'))
        self.date_entry.configure(font=('Helvetica', 12, 'bold'))
        self.time_entry.configure(font=('Helvetica', 12, 'bold'))

        # make labels helvetica, 12, bold
        self.dr_latitude_label.configure(font=('Helvetica', 12, 'bold'))
        self.dr_longitude_label.configure(font=('Helvetica', 12, 'bold'))
        self.date_label.configure(font=('Helvetica', 12, 'bold'))
        self.time_label.configure(font=('Helvetica', 12, 'bold'))

    
        # create buttons

        # make button text "Select a Time of Phenomena", and then when clicked, change text to "Plan For Selected Time"

        
        self.set_planning_time_button = ttk.Button(self.planning_controls_frame, text = 'Select a Time of Phenomena',
                                                        style='primary.Outline.TButton',
                                                       command= lambda: SightSessionPlanning(self.gui_page1.dr_entry_fields,
                                                                                             [self.time_of_phenomena_treeview,
                                                                                              self.planning_treeview,
                                                                                              self.optimal_triad_treeview
                                                                                              ],
                                                                                              [self.date_entry,
                                                                                               self.time_entry,
                                                                                               self.dr_latitude_entry,
                                                                                               self.dr_longitude_entry,
                                                                                               ]))
        
        # self.set_planning_time_button = ttk.Button(self.planning_controls_frame, text = 'Plan',
        #                                                 style='primary.Outline.TButton',
        #                                                command= lambda: SightSessionPlanning(self.gui_page1.dr_entry_fields,
        #                                                                                      [self.time_of_phenomena_treeview,
        #                                                                                       self.planning_treeview,
        #                                                                                       self.optimal_triad_treeview
        #                                                                                       ],
        #                                                                                       [self.date_entry,
        #                                                                                        self.time_entry,
        #                                                                                        self.dr_latitude_entry,
        #                                                                                        self.dr_longitude_entry,
        #                                                                                        ]))

        
        self.set_planning_time_button.grid(row=4, column=0, padx=10, pady=10, sticky='W', columnspan=7)


    def check(self):
        # if self.gui_page1.entry_fields are all complete, then instantiate SightSessionPlanning

        if all([entry.get() for entry in self.gui_page1.entry_fields[:5]]):
            # Instantiate SightSessionPlanning
            SightSessionPlanning(self.gui_page1.entry_fields,
                                 [self.time_of_phenomena_treeview,
                                  self.planning_treeview,
                                  self.optimal_triad_treeview],
                                 [self.date_entry,
                                  self.time_entry,
                                  self.dr_latitude_entry,
                                  self.dr_longitude_entry])
        else:
            Messagebox.show_error('Error', 'Please complete all fields')

    def create_time_of_phenomena_treeview(self):
        # add treeview to page 2
        self.time_of_phenomena_treeview = ttk.Treeview(self.phenoma_frame, style = 'danger.Treeview')
        
        # grid
        self.time_of_phenomena_treeview.grid(row=0, column=0, padx=10, pady=10, sticky='NESW')

        # Configure the weights
        self.phenoma_frame.grid_rowconfigure(0, weight=1)
        self.phenoma_frame.grid_columnconfigure(0, weight=1)

        
        # add date, time, event columns
        self.time_of_phenomena_treeview['columns'] = ('Date GMT', 'Date LMT', 'Event')


        # format columns
        self.time_of_phenomena_treeview.column('#0', width=0, stretch='no')
        self.time_of_phenomena_treeview.column('Date GMT', anchor='center', width=100)
        self.time_of_phenomena_treeview.column('Date LMT', anchor='center', width=100)
        self.time_of_phenomena_treeview.column('Event', anchor='center', width=100)

        # add headings
        self.time_of_phenomena_treeview.heading('#0', text='', anchor='w')
        self.time_of_phenomena_treeview.heading('Date GMT', text='Date GMT', anchor='center')
        self.time_of_phenomena_treeview.heading('Date LMT', text='Date LMT', anchor='center')
        self.time_of_phenomena_treeview.heading('Event', text='Event', anchor='center')

        # if the user selects row by clicking or scrolling, call the auto_fill_data_time function

        self.time_of_phenomena_treeview.bind('<ButtonRelease-1>', lambda event: self.auto_fill_date_time_entries())
        self.time_of_phenomena_treeview.bind('<KeyRelease>', lambda event: self.auto_fill_date_time_entries())

    def create_planning_treeview(self):
        # add treeview to page 3 
        self.planning_treeview = ttk.Treeview(self.page3)
        
        # grid
        self.planning_treeview.grid(row=0, column=0, padx=10, pady=10, sticky='NESW')

        # add body, altitude, and azimuth columns
        self.planning_treeview['columns'] = ('Body', 'Altitude', 'Azimuth', 'Magnitude')

        # format columns
        self.planning_treeview.column('#0', width=0, stretch='no')
        self.planning_treeview.column('Body', anchor='center', width=100)
        self.planning_treeview.column('Altitude', anchor='center', width=100)
        self.planning_treeview.column('Azimuth', anchor='center', width=100)
        self.planning_treeview.column('Magnitude', anchor='center', width=100)

        # add headings
        self.planning_treeview.heading('#0', text='', anchor='w')
        self.planning_treeview.heading('Body', text='Body', anchor='center')
        self.planning_treeview.heading('Altitude', text='Altitude', anchor='center')
        self.planning_treeview.heading('Azimuth', text='Azimuth', anchor='center')
        self.planning_treeview.heading('Magnitude', text='Magnitude', anchor='center')


    def create_optimal_triad_treeview(self):
        # add treeview to page 4
        self.optimal_triad_treeview = ttk.Treeview(self.page4)
        
        # grid
        self.optimal_triad_treeview.grid(row=0, column=0, padx=10, pady=10, sticky='NESW')

        # add body, altitude, azimuth, and magnitude columns
        self.optimal_triad_treeview['columns'] = ('Body', 'Altitude', 'Azimuth', 'Magnitude')

        # format columns
        self.optimal_triad_treeview.column('#0', width=0, stretch='no')
        self.optimal_triad_treeview.column('Body', anchor='center', width=100)
        self.optimal_triad_treeview.column('Altitude', anchor='center', width=100)
        self.optimal_triad_treeview.column('Azimuth', anchor='center', width=100)
        self.optimal_triad_treeview.column('Magnitude', anchor='center', width=100)

        # add headings
        self.optimal_triad_treeview.heading('#0', text='', anchor='w')
        self.optimal_triad_treeview.heading('Body', text='Body', anchor='center')
        self.optimal_triad_treeview.heading('Altitude', text='Altitude', anchor='center')
        self.optimal_triad_treeview.heading('Azimuth', text='Azimuth', anchor='center')
        self.optimal_triad_treeview.heading('Magnitude', text='Magnitude', anchor='center')

       

    def auto_fill_date_time_entries(self):
        # get date and time from sight planning treeview and fill in date and time entries

        # get selected item from treeview
        selected_item = self.time_of_phenomena_treeview.focus()

        # get values from selected item
        values = self.time_of_phenomena_treeview.item(selected_item, 'values')

        # fill in date and time entries with selected values
        self.date_entry.delete(0, 'end')
        
        # parse date and time from value
        date = values[0].split(' ')[0]

        # fill in date entry
        self.date_entry.insert(0, date)

        # fill in time entry
        self.time_entry.delete(0, 'end')
        self.time_entry.insert(0, values[0].split(' ')[1])

        # after filling time entry, if it is missing values in the seconds, add :00, else leave it alone
        if len(self.time_entry.get()) < 8:
            self.time_entry.insert(6, ':00')
        
        return 
    
    def autocompletion_binding(self):
        # initialize autocompletion
        self.autocomplete = AutoComplete(self.master)

        # bind autocompletion to date and time entries
        self.date_entry.bind('<KeyRelease>', lambda event: self.autocomplete.date_formatting(event, self.date_entry))
        self.time_entry.bind('<KeyRelease>', lambda event: self.autocomplete.time_formatting(event, self.time_entry))

    def create_tooltips(self):

        # Get the directory where the current script (presumably __main__.py or similar) is located
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # go up one directory from current_dir
        script_dir = os.path.dirname(current_dir)
        
        # Construct the absolute path to tooltips.txt
        tooltips_path = os.path.join(script_dir, 'text_files', 'tooltips.txt')

        self.extractor = TextExtractor(tooltips_path) 

        # tooltip for notebook page 1
        page1_tooltip = self.extractor.get_text('sight_planning')
        ToolTip(self.page1, text=page1_tooltip)

        # tooltip for sight planning treeview
        planning_tooltip = self.extractor.get_text('visible_bodies')
        ToolTip(self.planning_treeview, text=planning_tooltip)

        # tooltip for optimal triad treeview
        optimal_tooltip = self.extractor.get_text('sorted_triads')
        ToolTip(self.optimal_triad_treeview, text=optimal_tooltip)

        # tooltip for button
        button_tooltip = self.extractor.get_text('sight_planning')
        ToolTip(self.planning_controls_frame, text=button_tooltip)

        # tooltip for time of phenomena treeview
        time_of_phenomena_tooltip = self.extractor.get_text('time_of_phenomena')
        ToolTip(self.time_of_phenomena_treeview, text=time_of_phenomena_tooltip)


import utilities.celestial_engine as cnav
import tkinter as tk
import ttkbootstrap as ttk
from ttkbootstrap.dialogs import Messagebox
from collections import Counter
import datetime as dt
from tabulate import tabulate
import pyperclip as pc

class SightSessionPlanning():
    def __init__(self, ents, treeviews, fields):
    
        # page 1 date, time, latitude, longitude, course, speed entries
        self.ent_date = ents[0].get()
        self.ent_time = ents[1].get()
        self.ent_latitude = ents[2].get()
        self.ent_longitude = ents[3].get()
        self.ent_course = ents[4].get()
        self.ent_speed = ents[5].get()

        # page 4 date, time, latitude, longitude entry fields
        self.dr_date_entry = fields[0]
        self.dr_time_entry = fields[1]
        self.dr_latitude_entry = fields[2]
        self.dr_longitude_entry = fields[3]

        # values 
        self.dr_date_entry_value = fields[0].get()
        self.dr_time_entry_value = fields[1].get()
        self.dr_latitude_entry_value = fields[2].get()
        self.dr_longitude_entry_value = fields[3].get()

        # page 4 treeviews
        self.treeviews = treeviews

        self.delete_treeviews()
        self.update_treeviews()
        self.dr_sight_planning()
        self.visible_bodies_sorted_triads(self.ent_time, self.latitude, self.longitude)
        self.print_planning_info()

    def delete_treeviews(self):
        # delete any values in treeviews
        for treeview in self.treeviews:
            treeview.delete(*treeview.get_children())
    
    def update_treeviews(self):
        
        try:
            # convert to dd-mm.t format 
            self.latitude = cnav.Utilities.hmt_str_to_decimal_d(self.ent_latitude, self.ent_longitude)[0]
            self.longitude = cnav.Utilities.hmt_str_to_decimal_d(self.ent_longitude, self.ent_longitude)[1]

            
            # get phenomena times
            self.phenomena_times = cnav.Utilities.time_of_phenomena(self.ent_date, 
                                                                    self.ent_time, 
                                                                    self.latitude, 
                                                                    self.longitude,
                                                                    self.ent_course, 
                                                                    self.ent_speed)
            # add times of phenomena to treeview
            for element in self.phenomena_times:
                # make helvetica, 10, bold
                self.treeviews[0].tag_configure('bold', font=('Helvetica', 10, 'bold'))
                self.treeviews[0].insert('', 'end', text='', iid = element, values=element, tags=('bold',))
        except:
            Messagebox.show_error('Make sure the DR values are entered on page 1!', title = 'Incomplete DR Data')

    
    def dr_sight_planning(self):
        
        # delete any values in DR entry fields
        self.dr_latitude_entry.delete(0, 'end')
        self.dr_longitude_entry.delete(0, 'end')

        # get time from DR page and entry from sight planning page
        self.datetime_dr_start = cnav.Utilities.datetime(self.ent_date, self.ent_time)
        self.datetime_dr_end = cnav.Utilities.datetime(self.dr_date_entry_value, self.dr_time_entry_value)

        # get timedelta in seconds
        self.timedelta = dt.timedelta.total_seconds(self.datetime_dr_end - self.datetime_dr_start)

        # format latitude and longitude
        self.dr_lat_start = cnav.Utilities.hmt_str_to_decimal_d(self.ent_latitude, self.ent_longitude)[0]
        self.dr_long_start = cnav.Utilities.hmt_str_to_decimal_d(self.ent_longitude, self.ent_longitude)[1]

        # calculate DR position
        self.dr_lat_end = cnav.DRCalc(self.dr_lat_start, self.dr_long_start, self.timedelta, float(self.ent_course), float(self.ent_speed)).drlatfwds
        self.dr_long_end = cnav.DRCalc(self.dr_lat_start, self.dr_long_start, self.timedelta, float(self.ent_course), float(self.ent_speed)).drlongfwds

        # insert DR position into entry fields
        self.dr_latitude_entry.insert(0, cnav.Utilities.print_position2(self.dr_lat_end, latitude=True))
        self.dr_longitude_entry.insert(0, cnav.Utilities.print_position2(self.dr_long_end, latitude=False))


            
    def visible_bodies_sorted_triads(self, datetime, latitude, longitude):
            """Function to find all visible celestial bodies at the time requested, and sort them into weighted
            triads based on azimuth and magnitude """

            datetime = cnav.Utilities.datetime(self.ent_date, self.ent_time)
            possibleobvs = []
            named_bodies = ['SunLL', 'SunUL', 'MoonLL', 'MoonUL', 'Mars', 'Venus', 'Jupiter', 'Saturn']
            named_stars = [*cnav.Sight.named_star_dict]
            options = named_bodies + named_stars

            for body in options:
                ephem = cnav.Utilities.get_gha_dec(body, datetime, latitude, longitude)
                obsv = (body, ephem[3].degrees, ephem[4].degrees, ephem[5])
                # constrain to visible bodies that are easily shot
                if 10 < obsv[1] < 65:
                    obsv = (body, cnav.Utilities.hmt_str(ephem[3].degrees), round(ephem[4].degrees), ephem[5])
                    possibleobvs.append(obsv)

            triads = []
            # iterative process to create triad groupings
            for i in range(len(possibleobvs)):
                one = possibleobvs[i]
                try:
                    for x in range(len(possibleobvs)):
                        difference = (possibleobvs[i][2] - possibleobvs[x][2]) % 360
                        if difference > 115 and difference < 130:
                            two = possibleobvs[x]
                            for y in range(len(possibleobvs)):
                                difference2 = (possibleobvs[x][2] - possibleobvs[y][2]) % 360
                                if difference2 > 115 and difference2 < 130:
                                    three = possibleobvs[y]
                                    triad = [one, two, three]
                                    triads.append(triad)
                except:
                    pass

            sorted_triads = sorted(triads, key=lambda x: (x[0][3] + x[1][3] + x[2][3]) / 3)
            triad_bodies = []
            for i in range(len(sorted_triads)):
                body_list = (sorted_triads[i][0][0], sorted_triads[i][1][0], sorted_triads[i][2][0])
                triad_bodies.append(body_list)

            seen = set()

            triad_results = []
            for lst in triads:
                current = frozenset(Counter(lst).items())
                if current not in seen:
                    triad_results.append(lst)
                    seen.add(current)

            for i in possibleobvs:
                # make helvetica, 10, bold
                self.treeviews[1].tag_configure('bold', font=('Helvetica', 10, 'bold'))
                self.treeviews[1].insert('', 'end', text='', iid=i, values=i, tags=('bold',))
            counter = 0

            i = 1
            while i < len(triad_results):
                triad_results.insert(i, ('-'))
                i += (1 + 1)

            for i in triad_results:
                for y in i:
                    # make helvetica, 10, bold
                    self.treeviews[2].tag_configure('bold', font=('Helvetica', 10, 'bold'))
                    self.treeviews[2].insert('', 'end', text='', iid=counter, values=y, tags=('bold',))
                    counter += 1
            return
            
    def print_planning_info(self):
        """Aggregates all of the planning information and consolidates it into a printable format"""
        # get times of phenomena
        self.phenomena_times = cnav.Utilities.time_of_phenomena(self.ent_date, 
                                                                self.ent_time, 
                                                                self.latitude, 
                                                                self.longitude,
                                                                self.ent_course, 
                                                                self.ent_speed) 

        # make a markdown table using tabulate, get headers from treeviews[0]
        
        # get headers from treeview
        self.phenomena_headers = ['Date GMT', 'Date LMT', 'Event']

        # get values from treeview
        self.values = []
        for child in self.treeviews[0].get_children():
            self.values.append(self.treeviews[0].item(child)['values'])

        # create table
        self.table = tabulate(self.values, headers = self.phenomena_headers, tablefmt='github')

        # get visible bodies
        self.visible_bodies_headers = ['Body', 'Azimuth', 'Altitude', 'Magnitude']
        self.visible_bodies = []
        for child in self.treeviews[1].get_children():
            self.visible_bodies.append(self.treeviews[1].item(child)['values'])

        # create table
        self.table1 = tabulate(self.visible_bodies, headers = self.visible_bodies_headers, tablefmt='github')

        # get triads
        self.triad_headers = ['Body', 'Altitude', 'Azimuth', 'Magnitude'] 
        self.triads = []
        for child in self.treeviews[2].get_children():
            self.triads.append(self.treeviews[2].item(child)['values'])

        # create table
        self.table2 = tabulate(self.triads, headers = self.triad_headers, tablefmt='github')  

        # Header of planning information
        self.header = f"""\
        # Sight Planning Information
        ## Date: {self.ent_date}
        ## Time: {self.ent_time}
        ## Latitude: {self.ent_latitude}
        ## Longitude: {self.ent_longitude}
        ## Course: {self.ent_course}
        ## Speed: {self.ent_speed}
        """
        # copy to clipboard using pyperclip
        self.planning_info = self.header + '\n'*3 + self.table + '\n'*3 + self.table1 + '\n'*3 + self.table2
        pc.copy(self.planning_info)

        # show message box
        Messagebox.show_info('Planning information copied to clipboard!', title = 'Planning Information')

    
        




import tkinter as tk
import ttkbootstrap as ttk
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
import utilities.celestial_engine as celestial_engine

class LOPPlotPage(ttk.Frame):
    def __init__(self, container):
        super().__init__(container)

        self.add_lop_plot()
        self.draw_canvas()
        
    def add_lop_plot(self):
        # Create canvas
        self.canvas_lop = tk.Canvas(self)

        # Get figure from lop_plot 
        self.lop_plot = celestial_engine.plt.figure(2)
        self.lop_plot.set_facecolor('#222222')

        # Add figure to canvas
        self.lop_plot_canvas = FigureCanvasTkAgg(self.lop_plot, master=self.canvas_lop)

        # Add toolbar to canvas
        self.lop_plot_toolbar = NavigationToolbar2Tk(self.lop_plot_canvas, self.canvas_lop)
        self.lop_plot_toolbar.update()

        # Grid layout for toolbar, canvas, and plot canvas widget
        self.lop_plot_toolbar.pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady = 0)
        self.canvas_lop.pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady = 0)
        self.lop_plot_canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady = 0)

    def draw_canvas(self):
        # draw canvas
        self.lop_plot_canvas.draw()

    def refresh_figure(self):
        self.draw_canvas()





import tkinter as tk
import os
import ttkbootstrap as ttk

class SplashPage(ttk.Frame):
    def __init__(self, container):
        super().__init__(container)

        self.create_splash_page()

    import os

    def create_splash_page(self):
        # Get the directory where splash_page.py is located
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # Go one level up to the new_capella directory
        base_dir = os.path.dirname(current_dir)

    # Build the absolute path to splash_text.txt
        splash_text_path = os.path.join(base_dir, 'text_files', 'splash_text.txt')

        # Read splash_text.txt with utf-8 encoding
        with open(splash_text_path, 'r', encoding='utf-8') as f:
            self.splash_text = f.read()


        # make font courier 12
        large_font = ('Courier New', 12)

        logo = """
                     ______                 ____
                    / ____/___ _____  ___  / / /___ _
                   / /   / __ `/ __ \/ _ \/ / / __ `/
                  / /___/ /_/ / /_/ /  __/ / / /_/ /
 _________________\____/\__,_/ .___/\___/_/_/\__,_/_________________
/_____/_____/_____/         /_/                  /_____/_____/_____/

        """

        # make logo label
        self.logo_label = ttk.Label(self, text=logo, font=large_font, anchor = 'center', justify='left')


        # make label
        self.text_label = ttk.Label(self, text=self.splash_text, font=large_font, anchor = 'center', justify='center')

        # make it red
        self.logo_label.config(foreground='red')

        # make button
        self.button = ttk.Button(self, text='I ACCEPT THE ABOVE TERMS', command=self.enter, style='primary.Outline.TButton') 

        # grid everything
        self.logo_label.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.text_label.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.button.pack(side=tk.TOP, fill=tk.BOTH, padx=10, pady=10, expand=True)

    def enter(self):
        self.destroy()
    


import tkinter as tk
import ttkbootstrap as ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import datetime as dt
from ttkwidgets.autocomplete import AutocompleteCombobox
import utilities.celestial_engine as cnav
from ttkbootstrap.dialogs import Messagebox
from tkinter import simpledialog
import numpy as np
import geomag as gm
from skyfield.api import Angle, utc
from utilities.autocompletion import AutoComplete

class AzimuthPage(ttk.Frame):
    
    def __init__(self, container):
        super().__init__(container)
        self.init_widgets()

    def init_widgets(self):
        self.create_label_frames()
        self.create_observation_inputs()
        self.create_treeview()
        self.autocompletion_binding()

    def create_label_frames(self):
        self.azwrap = ttk.LabelFrame(self, text='Compass Observations Records')
        self.azwrap2 = ttk.LabelFrame(self, text='Observation Input')
        self.azwrap.pack(fill='both', expand=True, padx=10, pady=10)
        self.azwrap2.pack(expand=True, padx=10, pady=10)

        # self.azwrap.grid(row=0, column=0, sticky='NESW')
        # self.azwrap2.grid(row=1, column=0, sticky='NESW')

        # weights
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        

    def create_observation_inputs(self):
        # Define unique StringVar variables for each input
        self.az_vars = [tk.StringVar(self) for _ in range(12)]

        self.populate_dropdown()
        self.populate_entries()

        compute_btn = ttk.Button(self.azwrap2, text="COMPUTE", command=self.compass_correction, style='primary.Outline.TButton')
        compute_btn.grid(row=6, column=0, padx=10, pady=10)
        compute_btn.configure(width='12')

    def populate_dropdown(self):
        named_bodies = ['Pier Hdg', 'SunLL', 'SunUL', 'MoonLL', 'MoonUL', 'Mars', 'Venus', 'Jupiter', 'Saturn']
        named_stars = [*cnav.Sight.named_star_dict]
        options = named_bodies + named_stars

        azlbl1 = ttk.Label(self.azwrap2, text="Body", width=10, anchor='w')
        azlbl1.grid(row=0, column=0, padx=10, pady=10, sticky='w')
        
        self.azbodies = AutocompleteCombobox(self.azwrap2, textvariable=self.az_vars[0], completevalues=options, width=12)
        self.azbodies['values'] = options
        self.azbodies.grid(row=0, column=1, padx=10, pady=10)

        # make AutoComplete Combobox helvetica, 10, bold
        azlbl1.configure(font=('Helvetica', 10, 'bold'))
        self.azbodies.configure(font=('Helvetica', 10, 'bold'))

        # Bind the event for value change
        self.azbodies.bind("<<ComboboxSelected>>", self.on_combobox_selected)

    def on_combobox_selected(self, event):
        selected_value = self.azbodies.get()
        if selected_value == "Pier Hdg":
            heading = simpledialog.askfloat("Input", "Enter the heading:")
            if heading is not None:
                # Set the heading value to the combobox
                self.azbodies.set(heading)
            else:
                Messagebox.showwarning("Warning", "No heading entered.")
        
    def populate_entries(self):
        labels_texts = ["Gryo Hd", "Mag Hd", "Gyro Obsv."]
        default_values = ['300', '300', '300']

        for idx, (label_text, default_value) in enumerate(zip(labels_texts, default_values)):
            lbl = ttk.Label(self.azwrap2, text=label_text, width=10, anchor='w')
            lbl.grid(row=idx + 1, column=0, padx=10, pady=10, sticky='w')

            ent = ttk.Entry(self.azwrap2, textvariable=self.az_vars[idx + 1], width=20)
            # make entries helvetica, 10, bold
            ent.configure(font=('Helvetica', 10, 'bold'))
            ent.insert(0, default_value)
            ent.grid(row=idx + 1, column=1, padx=10, pady=10)
            # make labels helvetica, 10, bold
            lbl.configure(font=('Helvetica', 10, 'bold'))

        date_label = ttk.Label(self.azwrap2, text="Date UTC", width=10, anchor='w')
        date_label.grid(row=0, column=2, padx=10, pady=10, sticky='w')
        self.date_entry = ttk.Entry(self.azwrap2, textvariable=self.az_vars[8], width=20)
        self.date_entry.grid(row=0, column=3, padx=10, pady=10)
        self.date_entry.insert(0, dt.datetime.utcnow().strftime('%Y-%m-%d'))

        time_label = ttk.Label(self.azwrap2, text="Time UTC", width=10, anchor='w')
        time_label.grid(row=1, column=2, padx=10, pady=10, sticky='w')
        self.time_entry = ttk.Entry(self.azwrap2, textvariable=self.az_vars[9], width=20)
        self.time_entry.grid(row=1, column=3, padx=5, pady=3)
        self.time_entry.insert(0, dt.datetime.utcnow().strftime('%H:%M:%S'))

        lat_label = ttk.Label(self.azwrap2, text="DR Lat", width=10, anchor='w')
        lat_label.grid(row=2, column=2, padx=10, pady=10, sticky='w')
        self.lat_entry = ttk.Entry(self.azwrap2, textvariable=self.az_vars[10], width=20)
        self.lat_entry.grid(row=2, column=3, padx=5, pady=3)

        lon_label = ttk.Label(self.azwrap2, text="DR Long", width=10, anchor='w')
        lon_label.grid(row=3, column=2, padx=10, pady=10, sticky='w')
        self.lon_entry = ttk.Entry(self.azwrap2, textvariable=self.az_vars[11], width=20)
        self.lon_entry.grid(row=3, column=3, padx=10, pady=10)

        # make labels helvetica, 10, bold
        date_label.configure(font=('Helvetica', 10, 'bold'))
        time_label.configure(font=('Helvetica', 10, 'bold'))
        lat_label.configure(font=('Helvetica', 10, 'bold'))
        lon_label.configure(font=('Helvetica', 10, 'bold'))
        

        # make entries helvetica, 10, bold  
        self.date_entry.configure(font=('Helvetica', 10, 'bold'))
        self.time_entry.configure(font=('Helvetica', 10, 'bold'))
        self.lat_entry.configure(font=('Helvetica', 10, 'bold'))
        self.lon_entry.configure(font=('Helvetica', 10, 'bold'))
        
        # make autocomplete combobox helvetica, 10
        self.azbodies.configure(font=('Helvetica', 10))



    def create_treeview(self):
        self.trvaz = ttk.Treeview(self.azwrap, show='headings', height='12')
        self.trvaz.pack(side=tk.TOP, fill=tk.X, padx=10, pady=10)
        
        self.trvaz['columns'] = [str(i) for i in range(1, 14)]
        
        columns_width = [69, 60, 62, 66, 47, 48, 48, 48, 48, 54, 54, 54, 50]
        for col, width in zip(self.trvaz['columns'], columns_width):
            self.trvaz.column(col, anchor='center', width=width)

        headings = ["Date", 'G.M.T.', 'L', "λ", "G Hd", "C Hd", "T Brg", "G Brg", "Obj", "G Err", "C Err", "Var", "Dev"]
        for col, heading in zip(self.trvaz['columns'], headings):
            self.trvaz.heading(col, text=heading)

    def compass_correction(self):
        year, month, day = self.az_vars[8].get().split('-')
        hour, minute, second = self.az_vars[9].get().split(':')
        datetimeaz = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), tzinfo=utc)

        latitude = self.parse_lat_lon(self.az_vars[10].get())
        longitude = self.parse_lat_lon(self.az_vars[11].get())

        body = self.az_vars[0].get()

        try:
            az = float(body)  # Try to convert body to a float (to check if it's a Pier Heading number)
        except ValueError:
            if body == 'Pier Hdg':
                pier_heading = simpledialog.askfloat("Input", "Enter Pier Heading:")
                if pier_heading is not None:
                    self.azbodies.set(pier_heading)  # Update the combobox value
                    az = pier_heading
                else:
                    Messagebox.showwarning('Input Error', 'Enter Pier Heading')
                    return
            else:
                ephem = cnav.Utilities.get_gha_dec(body, datetimeaz, latitude, longitude)
                az = ephem[4].degrees

        gyro_hd = float(self.az_vars[1].get())
        std_hd = float(self.az_vars[2].get())
        gyro_az = float(self.az_vars[3].get())
        variation = np.round(float(gm.declination(latitude, longitude, 0)), 1)

        gyro_error = np.round(float(gyro_az - az), 1) if body != 'Pier Hdg' else np.round(float(gyro_az - az), 1)
        gyro_error_str = self.format_direction(gyro_error, 'E', 'W')
        deviation = np.round(gyro_hd - gyro_error - variation - std_hd, 1)
        com_error = np.round(variation + deviation, 1)

        az_date = datetimeaz.strftime("%Y-%m-%d")
        az_time = datetimeaz.strftime("%H:%M:%S")
        az_lat = cnav.Utilities.print_position(latitude, latitude=True)
        az_lon = cnav.Utilities.print_position(longitude, latitude=False)
        var_str = self.format_direction(variation)
        dev_str = self.format_direction(deviation)
        com_error_str = self.format_direction(com_error)

        for i in self.trvaz.get_children():
            self.trvaz.delete(i)
        trvazstr = [az_date, az_time, az_lat, az_lon, gyro_hd, std_hd, np.round(az, 1) if body != 'Pier Hdg' else np.round(az, 1), gyro_az, body, gyro_error_str, com_error_str, var_str, dev_str]

        # small font 
        self.trvaz.tag_configure('small', font=('Verdana', 7))
        self.trvaz.insert('', 'end', text='', iid=0, values=trvazstr, tags=('small',))

    def parse_lat_lon(self, value):
        deg, minutes, direction = value.split('-')
        return (float(deg) + (float(minutes) / 60)) * (-1 if direction in ['S', 'W'] else 1)

    def format_direction(self, value, positive='E', negative='W'):
        return f'{abs(value)} {positive if value >= 0 else negative}'
    
    def autocompletion_binding(self):
        # initialize the autocomplete
        self.autocomplete = AutoComplete(self.master)

        self.lat_entry.bind("<KeyRelease>", lambda event: self.autocomplete.lat_formatting(event, self.lat_entry))

        self.lon_entry.bind("<KeyRelease>", lambda event: self.autocomplete.long_formatting(event, self.lon_entry))

        self.time_entry.bind("<KeyRelease>", lambda event: self.autocomplete.time_formatting(event, self.time_entry))

        self.date_entry.bind("<KeyRelease>", lambda event: self.autocomplete.date_formatting(event, self.date_entry))




import re 

class InputChecking:
    """Class for checking input data"""

    def __init__(self, master):
        """Initializes the class"""
        self.master = master

    def check_time_format(x) -> bool:
        """Checks for hh:mm:ss format"""
        import re
        pattern = r'^([0-1]?\d|2[0-3])(?::([0-5]?\d))?(?::([0-5]?\d))?$'
        match = re.search(pattern, x)
        if match:

            return True
        else:
            return False

    def check_date_format(x) -> bool:
        """Checks for yyyy-mm-dd format"""
        import re

        pattern = r'^[0-9]{4}-(((0[13578]|(10|12))-(0[1-9]|[1-2][0-9]|3[0-1]))|(02-(0[1-9]|[1-2][0-9]))|((0[469]|11)-(0[1-9]|[1-2][0-9]|30)))$'
        match = re.search(pattern, x)
        if match:
            return True
        else:
            return False

    def check_hs_format(x) -> bool:
        """Checks for dd-mm.t format"""
        pattern = r"^([0-8][0-9]|89)+-(0?[0-9]|[1-5][0-9])\.\d"
        match = re.search(pattern, x)
        if match:

            return True
        else:
            return False

    def check_lat_format(x) -> bool:
        """Checks for dd-mm.t-N/S format"""
        pattern = r"^([0-8][0-9]|89)+-(0?[0-9]|[1-5][0-9])\.\d-[N|S]+"
        match = re.search(pattern, x)
        if match:
            return True
        else:
            return False

    def check_long_format(x) -> bool:
        """Checks for ddd-mm.t-E/W format"""
        pattern = r"^([0-1][0-9][0-9]|179)+-(0?[0-9]|[1-5][0-9])\.\d-[W|E]+"
        match = re.search(pattern, x)
        if match:
            return True
        else:
            return False

    def validate_number(x) -> bool:
        """Validates that the input is a number"""
        if x.strip('-').replace('.', '').isdigit():
            return True
        elif x == "":
            return True
        else:
            return False
        
    def check_celestial_body(x) -> bool:
        """Checks that the body is in the list of celestial bodies"""
        
        named_star_dict = {
            'Acamar': 13847, 'Achernar': 7588, 'Acrux': 60718, 'Adhara': 33579, 'Aldebaran': 21421, 'Algol': 14576,
            'Alioth': 62956,
            'Alkaid': 67301, 'Alnair': 109268, 'Alnilam': 26311, 'Alphard': 46390, 'Alphecca': 76267, 'Alpheratz': 677,
            'Altair': 97649,
            'Ankaa': 2081, 'Antares': 80763, 'Arcturus': 69673, 'Atria': 82273, 'Avior': 41037, 'Becrux': 62434,
            'Bellatrix': 25336,
            'Betelgeuse': 27989, 'Canopus': 30438, 'Capella': 24608, 'Deneb': 102098, 'Denebola': 57632, 'Diphda': 3419,
            'Dubhe': 54061,
            'Elnath': 25428, 'Enif': 107315, 'Eltanin': 87833, 'Fomalhaut': 113368, 'Gacrux': 61084, 'Gienah': 102488,
            'Hadar': 68702, 'Hamal': 9884, 'Kaus Australis': 90185, 'Kochab': 72607, 'Markab': 113963, 'Menkent': 68933,
            'Merak': 53910,
            'Miaplacidus': 45238, 'Mirach': 5447, 'Mirfak': 15863, 'Nunki': 92855, 'Peacock': 100751, 'Polaris': 11767,
            'Pollux': 37826,
            'Procyon': 37279, 'Rasalhague': 86032, 'Regulus': 49669, 'Rigel': 24436, 'Rigel Kent': 71683,
            'RigilKentaurus': 71683,
            'Sabik': 84012, 'Schedar': 3179, 'Shaula': 85927, 'Sirius': 32349, 'Spica': 65474, 'Suhail': 44816,
            'Vega': 91262,
            'Zubenelgenubi': 72622,
        }

        planets = ['SunLL', 'SunUL',  'MoonLL', 'MoonUL' 'Mercury', 'Venus', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune']
        if x in named_star_dict.keys() or x in planets:
            return True
        else:
            return False




import sys
import os
import subprocess
import numpy as np
from skyfield.api import Angle
import pyperclip as pc
import re
import ttkbootstrap as ttk
from ttkbootstrap.dialogs import Messagebox
import datetime as dt
import utilities.celestial_engine as celestial_engine
from utilities.sight_planning import SightSessionPlanning
from utilities.input_checking import InputChecking
from tabulate import tabulate

def save_sights_to_clipboard(instance, entries, sight_list_treeview):
    """
    Saves Sight Session and Sight info from Sight List trv and Session Data section, 
    formats it as a Markdown table and saves to clipboard.
    """
    session_array = []
    sight_array = []

    # Retrieve session data
    session = [ent.get() for ent in entries]
    session_array.append(session)

    # Retrieve sight data
    for record in sight_list_treeview.get_children():
        sight = sight_list_treeview.item(record, 'values')
        sight_array.append(sight)

    # Create Markdown tables with headers
    session_headers = ["DR Date", "DR Time", "DR L", "DR λ", "Course", "Speed"]
    sextant_headers = ["I.C.", "H.O.E", "Temp.", "Press."]
    fix_headers = ["Fix Date", "Fix Time"]
    sight_headers = ["Body", "Hs", "Date", "Time"]

    # Split session data into separate sections
    dr_data = session_array[0][:6]
    sextant_data = [session_array[0][6:10]]
    fix_data = [session_array[0][10:]]

    # Format each section with headers
    dr_copy = "### 1. Dead Reckoning Information\n" + tabulate([dr_data], headers=session_headers, tablefmt="github")
    sextant_copy = "### 2. Sextant Information\n" + tabulate(sextant_data, headers=sextant_headers, tablefmt="github")
    fix_copy = "### 3. Fix Date and Time Information\n" + tabulate(fix_data, headers=fix_headers, tablefmt="github")
    sight_copy = "### 4. Sights\n" + tabulate(sight_array, headers=sight_headers, tablefmt="github")

    # Combine all data
    copied_data = "\n\n".join([dr_copy, sextant_copy, fix_copy, sight_copy])
    pc.copy(copied_data)

    return dr_copy, sextant_copy, fix_copy, sight_copy

def load_sights_from_clipboard(instance, entries, sight_list_treeview):
    """
    Loads Sight Session DR info and Sights into the Session info Sights Treeview from the clipboard.
    """
    copied_text = pc.paste()

    try:
        # Split the copied text into sections
        sections = re.split(r'### \d\.\s.*\n', copied_text)
        sections = [section for section in sections if section.strip()]

        # Process DR information (section 0)
        dr_rows = sections[0].strip().split('\n')[2:]  # Skip headers and divider rows
        dr_data = [cell.strip() for cell in dr_rows[0].split('|') if cell.strip()]
        for j, value in enumerate(dr_data):
            entries[j].delete(0, 'end')
            entries[j].insert(0, value)

        # Process Sextant information and Fix Date and Time (sections 1 and 2)
        for i in range(1, 3):
            rows = sections[i].strip().split('\n')[2:]  # Skip headers and divider rows
            data = [cell.strip() for cell in rows[0].split('|') if cell.strip()]
            for j, value in enumerate(data):
                entry_index = 6 + (i - 1) * 4 + j  # Adjust index based on section
                entries[entry_index].delete(0, 'end')
                entries[entry_index].insert(0, value)

        # Clear Sight Entry treeview
        for i in sight_list_treeview.get_children():
            sight_list_treeview.delete(i)

        # Populate Sight Entry treeview (section 3)
        if len(sections) > 3:
            sights = sections[3].strip().split('\n')[2:]  # Skip headers and divider rows
            for i, sight in enumerate(sights):
                sight_data = [cell.strip() for cell in sight.split('|') if cell.strip()]
                sight_list_treeview.tag_configure('main', font=('Arial Bold', 15))
                sight_list_treeview.insert('', 'end', text='', iid=i, 
                                           values=sight_data, tags=('main',))
                instance.counter += 1   
    except:
        Messagebox.show_warning(title = 'Input Error', message = 'Data not in recognized format, check clipboard data.')
        return
    
    # check sight info for errors
    for i, record in enumerate(sight_list_treeview.get_children()):
        sight = sight_list_treeview.item(record, 'values')
        if not InputChecking.check_celestial_body(sight[0]):
            Messagebox.show_warning(title = f'Input Error Sight # {i+1}', message = 'Celestial Body Formatted Incorrectly, check entry in Sight Entry Treeview')
            return
        if not InputChecking.check_hs_format(sight[1]):
            Messagebox.show_warning(title = f'Input Error Sight # {i+1}', message = 'Hs Formatted Incorrectly, check entry in Sight Entry Treeview')
            return
        if not InputChecking.check_date_format(sight[2]):
            Messagebox.show_warning(title = f'Input Error Sight # {i+1}', message = 'Date Formatted Incorrectly, check entry in Sight Entry Treeview')
            return
        if not InputChecking.check_time_format(sight[3]):
            Messagebox.show_warning(title = f'Input Error Sight # {i+1}', message = 'Time Formatted Incorrectly, check entry in Sight Entry Treeview')
            return
        

def add_new_sight(instance, bodies_entry_box, entry_boxes, sight_list_treeview):
    
    """Adds a new row to the Sight Entry Treeview"""
    try:
        # Get values from entry boxes and add to Treeview
        values = [entry.get() for entry in entry_boxes]

        sight_list_treeview.tag_configure('main', font=('Arial Bold', 15))
        sight_list_treeview.insert('', 'end', text='', iid=instance.counter, values=values, tags=('main',))
        
        # Clear entry boxes
        for entry in entry_boxes:
            entry.delete(0, 'end')
        instance.counter += 1
        
    except Exception as e:
        print(f"Error adding new row to Sight Entry Treeview: {e}")
    
    # Set focus back to bodies autocomplete box
    bodies_entry_box.focus()

    return 

def delete_sight(sight_list_treeview):
    """Deletes selected row from Sight Entry Treeview"""
    selection = sight_list_treeview.selection()
    for record in selection:
        sight_list_treeview.delete(record)

def update_sight(entry_list, sight_list_treeview):
    """Updates entry fields in 'Sight Entry' section"""
    selected = sight_list_treeview.focus()
    selection = sight_list_treeview.item(selected, 'values')
    sight_list_treeview.tag_configure('main', font=('Arial Bold', 15))
    sight_list_treeview.item(selected, text='', values=(entry_list[0].get(), 
                                                        entry_list[1].get(), 
                                                        entry_list[2].get(), 
                                                        entry_list[3].get()), 
                                                        tags=('main', 0))


def open_sight_log(event=None):
    """Opens sight_log.txt file in the default text editor, in an OS-agnostic way."""
    # Define the path to the file
     # Get the directory where the current script (presumably __main__.py or similar) is located
    current_dir = os.path.dirname(os.path.abspath(__file__))

    # go down one level to the root directory
    root_dir = os.path.dirname(current_dir)

    # go down one more level to the text_files directory
    text_files_dir = os.path.join(root_dir, 'text_files')

    # Define the path to the file
    file_path = os.path.join(text_files_dir, 'sight_log.txt')
   

    # Check if the file exists
    if not os.path.exists(file_path):
        print(f"File not found: {file_path}")
        return

    # Open the file with the default application
    try:
        if sys.platform == 'win32':
            os.startfile(file_path)  # For Windows
        elif sys.platform == 'darwin':
            subprocess.Popen(['open', file_path])  # For macOS
        else:
            subprocess.Popen(['xdg-open', file_path])  # For Linux
    except Exception as e:
        print(f"Error opening file: {e}")    

class UpdateAndAveraging:
    def __init__(self, treeview, ents):
        self.treeview = treeview
        self.ents = ents

    def print_element(self, event):
        """Click on a Sight in the Sight Field treeview and the Sight Entry input box values will change
        respectively """
        trv = event.widget
        selected = trv.focus()
        selection = trv.item(selected, 'values')

        for ent in self.ents:
            ent.delete(0, 'end')

        self.ents[0].insert(0, selection[0])
        self.ents[1].insert(0, selection[1])
        self.ents[2].insert(0, selection[2])
        self.ents[3].insert(0, selection[3])

        # Sight Averaging
        selection = trv.selection()
        datetimeList = []
        hsList = []
        for record in selection:
            # time averaging
            values = trv.item(record, 'values')
            year, month, day = values[2].split('-')
            hour, minute, second = values[3].split(':')
            sight_dt_obj = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second))
            datetimeList.append(sight_dt_obj)
            avgTime = dt.datetime.strftime(dt.datetime.fromtimestamp(
                sum(map(dt.datetime.timestamp, datetimeList)) / len(datetimeList)), "%H:%M:%S")
            avgDate = dt.datetime.strftime(dt.datetime.fromtimestamp(
                sum(map(dt.datetime.timestamp, datetimeList)) / len(datetimeList)), "%Y-%m-%d")

            # hs averaging
            hs_deg, hs_min = values[1].split('-')
            hs = (float(hs_deg) + (float(hs_min) / 60))
            hs = Angle(degrees=(hs))
            hsList.append(hs.degrees)

            hs_avg = celestial_engine.Utilities.hmt_str_2(np.mean(hsList))

# make ent text red if more than one sight is selected

            if len(selection) >= 2:
                self.ents[1].config(foreground='cyan')
                self.ents[2].config(foreground='cyan')
                self.ents[3].config(foreground='cyan')
            else:
                self.ents[1].config(foreground='white')
                self.ents[2].config(foreground='white')
                self.ents[3].config(foreground='white')


            self.ents[1].delete(0, 'end')
            self.ents[2].delete(0, 'end')
            self.ents[3].delete(0, 'end')
            self.ents[1].insert(0, hs_avg)
            self.ents[2].insert(0, avgDate)
            self.ents[3].insert(0, avgTime)
            
            # if len(hsList) >= 2:
            #     avg_lbl.grid(row=1, column=2, padx=2, pady=3)
            #     avg_lbl_2.grid(row=3, column=2, padx=2, pady=3)
            # else:
            #     avg_lbl.grid_forget()
            #     avg_lbl_2.grid_forget()



import os
os.system('color')
from skyfield.api import load, Angle, Star, Topos, utc, wgs84
from skyfield import almanac
from skyfield.data import hipparcos
with load.open(hipparcos.URL) as f:
    df = hipparcos.load_dataframe(f)
import re, datetime as dt
from datetime import timezone, timedelta
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker
import matplotlib.dates as mdates

from matplotlib.patches import Ellipse
from tabulate import tabulate
import scipy.optimize as optimize

from utilities.dead_reckoning import DRCalc



class Utilities():

    def datetime(date, time):
        """ take user's date and time string values and concatenates into one datetime object.

        Parameters
        ----------
        date : str
            Always in UTC. 'yyyy-mm-dd'
        time : str
            Always in UTC. 'hh:mm:ss'

        """
        year, month, day = date.split('-')
        hour, minute, second = time.split(':')
        datetime = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), tzinfo=utc)

        return datetime

    def hms(time):
        """ enter a number in hh.mmss format and, it will return a hh.hhhh value, mimics the hp-48gx sexagesimal
            math functions.

        Parameters
        ----------
        time : float
            Always in UTC. hh.mmss returned as hh.hhhh
        """
        time_hours, time_minutes = divmod(time, 1)
        time_minutes = time_minutes * 100
        time_minutes, time_seconds = divmod(time_minutes, 1)
        time_seconds = time_seconds * 100
        time = float(time_hours + (time_minutes / 60) + (time_seconds / 3600))

        return time

    def hmt_str(angle):
        """ enter a skyfield angle object value in degrees, returns a str value formatted "dd°mm'".

        Parameters
        ----------
        angle : float in degrees, hh.hhhh format
        """
        deg = int(angle)
        min = float(np.round(abs(angle) % 1 * 60, 1))
        if min < 10:
            min = str(min).zfill(4)

        return f"{deg}°{min}'"

    def hmt_str_2(angle):
        """ enter a skyfield angle object value in degrees, returns a str value formatted "dd-mm".

        Parameters
        ----------
        angle : skyfield Angle object in degrees
        """
        deg = int(angle)
        min = float(np.round(abs(angle) % 1 * 60, 1))
        if min < 10:
            min = str(min).zfill(4)

        return f"{deg}-{min}"

    def hmt_str_to_decimal_d(latstr, longstr):
        """ convert latitude and longitude string values into float values

        Parameters
        ----------
        latstr : str
            latitude str value formatted 'dd-mm.t-N/S'

        longstr : str
            longitude str value formatted 'ddd-mm.t-E/W'

        """
        deg, minutes, direction = latstr.split('-')
        latitude = (float(deg) + (float(minutes) / 60)) * (-1 if direction in 'S' else 1)
        deg, minutes, direction = longstr.split('-')
        longitude = (float(deg) + (float(minutes) / 60)) * (-1 if direction in 'W' else 1)

        return latitude, longitude

    def hh_mm_ss(time):
        """ enter a number in hh.mmss format and, it will split up hh, mm, ss as a tuple.

        Parameters
        ----------
        time : float
            Always in UTC. hh.mmss format.
        """

        time_hours, time_minutes = divmod(time, 1)

        time_minutes = float(time_minutes * 100)
        time_minutes, time_seconds = divmod(time_minutes, 1)
        time_seconds = float(time_seconds * 100)

        timehours = float(time_hours)
        timeminutes = round(float(time_minutes), 1)
        timeseconds = round(float(time_seconds), 1)

        return timehours, timeminutes, timeseconds

    def hms_out(time):
        """ enter a number in hh.hhhh format and, it will return a number in hh.mmss format.

        Parameters
        ----------
        time : float
            Always in UTC. hh.hhh format.
        """

        time_real = time
        time_hours, time_minutes = divmod(abs(time), 1)
        time_minutes = (time_minutes * 60)
        timeminutes = time_minutes / 100

        time_minutes, time_seconds = divmod(time_minutes, 1)
        time_seconds = time_seconds * 60
        timeseconds = (time_seconds) / 100

        minutesseconds = round((time_minutes + timeseconds) / 100, 4)
        if time_real < 0:
            time = round(((time_hours + minutesseconds) * -1), 4)
        else:
            time = round((time_hours + minutesseconds), 4)

        return time

    def print_position(position, latitude=True):
        """ receives a float value latitude or longitude, adds N, S, E, W suffix based on type and value and converts
        to str value using hmt_str function.

        Parameters
        ----------
        position : float
            latitude or longitude in hh.hhhh format

        latitude : bool
            whether or, not the position is a longitude, determines the N, S, E, W suffix to use.
        """
        if latitude == True:
            if position > 0:
                sign = 'N'
                print_latitude = position
            else:
                sign = 'S'
                print_latitude = position * -1

            final_string = (f'{Utilities.hmt_str(print_latitude)} {sign}')
        if latitude != True:
            if position > 0:
                sign = "E"
                print_longitude = position
            else:
                sign = "W"
                print_longitude = position * -1

            final_string = (f'{Utilities.hmt_str(print_longitude)} {sign}')

        return final_string

    def print_position2(position, latitude=True):
        """ receives a float value latitude or longitude, adds N, S, E, W suffix based on type and value and converts
        to str value using hmt_str_2 function.

        Parameters
        ----------
        position : float
        latitude or longitude in hh.hhhh format

        latitude : bool
        whether or, not the position is a longitude determines the N, S, E, W suffix to use.
        """
        if latitude == True:
            if position > 0:
                sign = 'N'
                print_latitude = position
            else:
                sign = 'S'
                print_latitude = position * -1

            final_string = (f'{Utilities.hmt_str_2(print_latitude)}-{sign}')
        else:
            if position > 0:
                sign = "E"
                print_longitude = position
            else:
                sign = "W"
                print_longitude = position * -1

            final_string = (f'{Utilities.hmt_str_2(print_longitude)}-{sign}')

        return final_string

    def single_body_time_divide(obj_array):
        """ receives an array of single body sight tuples and splits them into buckets based on a 900-second (15 min)
            interval, it will then return the sight-time with the lowest d-value from each bucket. For example,
            6 shots of the sun with a group of 3 at 1000, and 3 at or around LAN is 2 sessions. It will return
            2 buckets of 3 sun shots with the lowest d scatter value per bucket.

        Parameters
        ----------
        obj_array : list of tuples
        each tuple is : ('object', index, d-value, datetime)
        """
        split_points = []
        sorted(split_points)

        for i in range(len(obj_array)):
            try:
                delta = (dt.timedelta.total_seconds(obj_array[i + 1][3] - obj_array[i][3]))

                if abs(delta) > 900: split_point = i
                split_points.append(split_point)
            except:
                pass

        split_points = set(split_points)

        unique_list_splits = (list(split_points))

        bucket1 = []
        bucket2 = []
        bucket3 = []
        finalbucket = []

        if len(split_points) == 0:
            for i in obj_array:
                bucket1.append((i, 'Bucket1'))
                bucket1dvals = []
                for i in bucket1:
                    bucket1dvals.append(i[0][2])
                sorted_valuesb1 = sorted(bucket1dvals, key=lambda x: abs(x))
            for i in bucket1:

                if i[0][2] == sorted_valuesb1[0]:
                    match1 = (i[0][0], i[0][1])
                    return [match1]

        if len(unique_list_splits) == 1:
            for i in obj_array:
                if obj_array.index(i) <= unique_list_splits[0]:
                    bucket1.append((i, 'Bucket1'))
                    bucket1dvals = []
                    for i in bucket1:
                        bucket1dvals.append(i[0][2])
                    sorted_valuesb1 = sorted(bucket1dvals, key=lambda x: abs(x))
                else:
                    bucket2.append((i, 'Bucket2'))
                    bucket2dvals = []
                    for i in bucket2:
                        bucket2dvals.append(i[0][2])
                    sorted_valuesb2 = sorted(bucket2dvals, key=lambda x: abs(x))
            for i in bucket1:
                if i[0][2] == sorted_valuesb1[0]:
                    match1 = (i[0][0], i[0][1])
            for i in bucket2:
                if i[0][2] == sorted_valuesb2[0]:
                    match2 = (i[0][0], i[0][1])
                    return match1, match2

        if len(unique_list_splits) == 2:
            for i in obj_array:
                if obj_array.index(i) <= unique_list_splits[0]:
                    bucket1.append((i, 'Bucket1'))
                    bucket1dvals = []
                    for i in bucket1:
                        bucket1dvals.append(i[0][2])
                    sorted_valuesb1 = sorted(bucket1dvals, key=lambda x: abs(x))
                elif obj_array.index(i) <= unique_list_splits[1]:
                    bucket2.append((i, 'Bucket2'))
                    bucket2dvals = []
                    for i in bucket2:
                        bucket2dvals.append(i[0][2])
                    sorted_valuesb2 = sorted(bucket2dvals, key=lambda x: abs(x))
                else:
                    bucket3.append((i, 'Bucket3'))
                    bucket3dvals = []
                    for i in bucket3:
                        bucket3dvals.append(i[0][2])
                    sorted_valuesb3 = sorted(bucket3dvals, key=lambda x: abs(x))

            for i in bucket1:
                if i[0][2] == sorted_valuesb1[0]:
                    match1 = (i[0][0], i[0][1])
            for i in bucket2:
                if i[0][2] == sorted_valuesb2[0]:
                    match2 = (i[0][0], i[0][1])
            for i in bucket3:
                if i[0][2] == sorted_valuesb3[0]:
                    match3 = (i[0][0], i[0][1])
                    return match1, match2, match3

        return

    def time_of_phenomena(date, time, dr_lat, dr_long, course, speed):
        """ receives date, time and dr information, calculates the time of phenomena for am/pm civil/nautical
        twilights and then DR's to that time to obtain a second estimate time.

        Parameters
        ----------
        date : str
            'yyyy-mm-dd'
        time : str
            'hh:mm:ss'
        dr_lat : float
            dd.dddd format, S is -
        dr_long : float
            dd.dddd format , W is -
        course : float
            ddd format
        speed : float
            dd format
        """

        year, month, day = date.split('-')
        hour, minute, second = time.split(':')

        zd = round(dr_long / 15)

        tz = timezone(timedelta(hours=zd))
        gmt = timezone(timedelta(hours=0))

        datetime = dt.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), tzinfo=tz)

        # Figure out local midnight.

        midnight = datetime.replace(hour=0, minute=0, second=0, microsecond=0)
        next_midnight = midnight + dt.timedelta(days=1)

        ts = load.timescale()
        t0 = ts.from_datetime(midnight)
        t1 = ts.from_datetime(next_midnight)
        eph = load('de421.bsp')
        position = wgs84.latlon(dr_lat, dr_long)
        f = almanac.dark_twilight_day(eph, position)
        f_1 = almanac.meridian_transits(eph, eph['Sun'], position)
        suntimes, sunevents = almanac.find_discrete(t0, t1, f_1)
        times, events = almanac.find_discrete(t0, t1, f)

        lan = suntimes[sunevents == 1]
        tsun = lan[0]
        sunstr = str(tsun.astimezone(tz))[:19]

        # second estimate for LAN

        sunstr = str(tsun.astimezone(tz))[:19]
        time_delta = dt.timedelta.total_seconds(tsun.astimezone(tz) - datetime)
        second_estimate_lat = DRCalc(dr_lat, dr_long, time_delta, course, speed).drlatfwds
        second_estimate_long = DRCalc(dr_lat, dr_long, time_delta, course, speed).drlongfwds
        position2 = wgs84.latlon(second_estimate_lat, second_estimate_long)
        f_1 = almanac.meridian_transits(eph, eph['Sun'], position2)
        suntimes, sunevents = almanac.find_discrete(t0, t1, f_1)
        lan = suntimes[sunevents == 1]
        tsun = lan[0]
        sunstr = str(tsun.astimezone(tz))[:19]
        sunstr2 = str(tsun.astimezone(gmt))[:19]

        # zd = round(second_estimate_long / 15)

        tz = timezone(timedelta(hours=zd))
        lanstr = (sunstr2, sunstr, 'L.A.N.')

        phenomenatimes = []

        previous_e = f(t0).item()
        for t, e in zip(times, events):
            tstr = str(t.astimezone(tz))[:16]
            tstr2 = str(t.astimezone(gmt))[:16]

            if previous_e < e:
                string = (tstr2, tstr, f'{almanac.TWILIGHTS[e]} starts [{zd}]')
                phenomenatimes.append(string)
                if len(phenomenatimes) == 4:
                    phenomenatimes.append(lanstr)

            else:
                string = (tstr2, tstr, f'{almanac.TWILIGHTS[previous_e]} ends [{zd}]')
                phenomenatimes.append(string)
                if len(phenomenatimes) == 4:
                    phenomenatimes.append(lanstr)

            previous_e = e

        return phenomenatimes

    def get_gha_dec(body, datetime, latitude, longitude):
        """ receives celestial object, date and position and returns gha, dec, ghaa, alt, az and magnitude using
        the Skyfield library, hipparcos catalog and DE421 database.

        Parameters
        ----------
        body : str
            Celestial object in question, upper limb and lower limb are UL or LL respectively.
        datetime: dt object

        latitude: float
            dd.dddd format, S is -

        longitude: float
            ddd.dddd format, W is -
        """

         # Dictionary for named stars and their hipparcos id
        named_star_dict = {
                'Acamar': 13847, 'Achernar': 7588, 'Acrux': 60718, 'Adhara': 33579, 'Aldebaran': 21421, 'Algol': 14576,
                'Alioth': 62956,
                'Alkaid': 67301, 'Alnair': 109268, 'Alnilam': 26311, 'Alphard': 46390, 'Alphecca': 76267, 'Alpheratz': 677,
                'Altair': 97649,
                'Ankaa': 2081, 'Antares': 80763, 'Arcturus': 69673, 'Atria': 82273, 'Avior': 41037, 'Becrux': 62434,
                'Bellatrix': 25336,
                'Betelgeuse': 27989, 'Canopus': 30438, 'Capella': 24608, 'Deneb': 102098, 'Denebola': 57632, 'Diphda': 3419,
                'Dubhe': 54061,
                'Elnath': 25428, 'Enif': 107315, 'Eltanin': 87833, 'Fomalhaut': 113368, 'Gacrux': 61084, 'Gienah': 102488,
                'Hadar': 68702, 'Hamal': 9884, 'Kaus Australis': 90185, 'Kochab': 72607, 'Markab': 113963, 'Menkent': 68933,
                'Merak': 53910,
                'Miaplacidus': 45238, 'Mirach': 5447, 'Mirfak': 15863, 'Nunki': 92855, 'Peacock': 100751, 'Polaris': 11767,
                'Pollux': 37826,
                'Procyon': 37279, 'Rasalhague': 86032, 'Regulus': 49669, 'Rigel': 24436, 'Rigel Kent': 71683,
                'RigilKentaurus': 71683,
                'Sabik': 84012, 'Schedar': 3179, 'Shaula': 85927, 'Sirius': 32349, 'Spica': 65474, 'Suhail': 44816,
                'Vega': 91262,
                'Zubenelgenubi': 72622,
            }

        planets = load('de421.bsp')
        ts = load.timescale()
        t = ts.utc(datetime)

        if body == 'SunLL' or body == 'SunUL':
            celestial_body = planets['Sun']
            mag = -26.74
        elif body == 'MoonLL' or body == 'MoonUL':
            celestial_body = planets['Moon']
            mag = - 12.6
        elif body == 'Mars':
            celestial_body = planets['Mars']
            mag = 1.4
        elif body == 'Venus':
            celestial_body = planets['Venus']
            mag = -4.9
        elif body == 'Jupiter':
            celestial_body = planets['Jupiter Barycenter']
            mag = -2.9
        elif body == 'Saturn':
            celestial_body = planets['Saturn Barycenter']
            mag = .75
        elif body == 'Uranus':
            celestial_body = planets['Uranus Barycenter']
            mag = 5.38
        elif body == 'Mercury':
            celestial_body = planets['Mercury']
            mag = .28
        else:
            which_star = body
            hid = named_star_dict.get(which_star)
            celestial_body = Star.from_dataframe(df.loc[hid])
            mag = df['magnitude'][hid]

        obs = planets['Earth']
        position = obs + Topos(latitude_degrees=(latitude), longitude_degrees=(longitude))
        astro = position.at(t).observe(celestial_body)
        app = astro.apparent()
        astrometric = obs.at(t).observe(celestial_body)
        apparent = obs.at(t).observe(celestial_body).apparent()
        alt, az, distance = app.altaz()
        ra, dec, distance, = apparent.radec(epoch='date')
        ghaa = Angle(degrees=(t.gast) * 15)
        gha = Angle(degrees=((t.gast - ra.hours) * 15 % 360 - 0))

        return gha, dec, ghaa, alt, az, mag

    def plot_cov_ellipse(cov, pos, nstd=2, ax=None, **kwargs):
        """ Generates confidence ellipse in matplot lib.

        Parameters
        ----------
        cov : np.array
            Covariance matrix from L-BFGS-B function
        pos : tuple
            x, y position for center of ellipse
        nstd : int
            number of standard deviations.
        """
        def eigsorted(cov):
            vals, vecs = np.linalg.eigh(cov)
            order = vals.argsort()[::-1]
            return vals[order], vecs[:, order]

        if ax is None:
            ax = plt.gca()

        vals, vecs = eigsorted(cov)

        theta = np.degrees(np.arctan2(*vecs[:, 0][::-1])) * -1

        # Width and height are "full" widths, not radius
        width, height = 2 * nstd * np.sqrt(vals)

        ellip = Ellipse(xy=pos, width=height / 100, height=width / 100, angle=theta, **kwargs)

        ax.add_artist(ellip)
        return ellip







import tkinter as tk
import os
import ttkbootstrap as ttk
from ttkbootstrap.tooltip import ToolTip
from ttkbootstrap.widgets import Floodgauge
from ttkwidgets.autocomplete import AutocompleteCombobox
import utilities.celestial_engine as cnav
from utilities.sight_handling import add_new_sight, delete_sight, update_sight, UpdateAndAveraging
from utilities.reduce_sight import CapellaSightReduction
from utilities.input_checking import InputChecking
from utilities.autocompletion import AutoComplete
from utilities.sight_planning import SightSessionPlanning
from utilities.tooltips import TextExtractor

class SightEntryPage(ttk.Frame):
    
    # sight treeview class variable
    counter = 0
    def __init__(self, container):
        super().__init__(container)

        self.create_label_frames()
        self.create_notebook()
        self.create_sight_treeview()
        self.create_dr_info_entry()
        self.create_sextant_info_entry()
        self.create_sight_info_entry()
        self.create_fix_info_entry()
        self.aggregate_entry_fields()
        self.autocompletion_binding()
        self.create_tooltips()

        # binds 
        self.sight_list_treeview.bind('<<TreeviewSelect>>', 
                                      lambda event: UpdateAndAveraging(self.sight_list_treeview,
                                                                       self.sight_entry_fields).print_element(event))
        self.sight_list_treeview.bind('<<TreeviewSelect>>', lambda event: (
            self.notebook.select(self.notebook.tabs()[2]),
            UpdateAndAveraging(self.sight_list_treeview, self.sight_entry_fields).print_element(event)
        ))
        
  
    def create_label_frames(self):

        self.sight_frame = ttk.LabelFrame(self, text='Sight List')
        self.sight_frame.pack(expand = True, fill = "both")

    def create_notebook(self):

        # create label frame
        self.notebook_frame = ttk.LabelFrame(self, text = 'Session Menu')
        self.notebook_frame.pack(expand = True, fill = 'both')
        self.notebook = ttk.Notebook(self.notebook_frame)
        
        # grid
        self.notebook.pack(expand = True, fill = "both") 
                
        self.entry_page_1 = ttk.Frame(self.notebook)
        self.entry_page_1.pack(expand = False)
        self.dr_info_frame = self.entry_page_1

        self.entry_page_2 = ttk.Frame(self.notebook)
        self.entry_page_2.pack(expand = False )
        self.sextant_info_frame = self.entry_page_2

        self.entry_page_3 = ttk.Frame(self.notebook)
        self.entry_page_3.pack(expand = False)
        self.sight_info_entry_frame = self.entry_page_3

        self.entry_page_4 = ttk.Frame(self.notebook)
        self.entry_page_4.pack(expand = False)
        self.fix_info_frame = self.entry_page_4

        self.notebook.add(self.entry_page_1, text='DR Info.')
        self.notebook.add(self.entry_page_2, text='Sextant Info.')
        self.notebook.add(self.entry_page_3, text = 'Sight Entry')
        self.notebook.add(self.entry_page_4, text='Fix Computation')
       
    def create_sight_treeview(self):
        # Configure the style for the treeview
        style = ttk.Style()
        style.configure('info.Treeview', rowheight=50) 

        # create treeview
        self.sight_list_treeview = ttk.Treeview(self.sight_frame, height=10, style = 'info.Treeview')

        # With these lines:
        self.sight_frame.grid_rowconfigure(0, weight=1)
        self.sight_frame.grid_columnconfigure(0, weight=1)
        self.sight_list_treeview.grid(row=0, column=0, ipadx=10, ipady=10, sticky='nsew')
                
        # add columns to treeview
        self.sight_list_treeview['columns'] = ('Body', 'Hs', 'Date', 'Time')
        self.sight_list_treeview.column('#0', width=0, stretch='no')
        self.sight_list_treeview.column('Body', anchor='center', width=140)
        self.sight_list_treeview.column('Hs', anchor='center', width=140)
        self.sight_list_treeview.column('Date', anchor='center', width=140)
        self.sight_list_treeview.column('Time', anchor='center', width=140)
        
        # add headings to treeview
        self.sight_list_treeview.heading('#0', text='', anchor='w')
        self.sight_list_treeview.heading('Body', text='Body', anchor='center')
        self.sight_list_treeview.heading('Hs', text='Hs', anchor='center')
        self.sight_list_treeview.heading('Date', text='Date', anchor='center')
        self.sight_list_treeview.heading('Time', text='Time', anchor='center')

                       

    def create_dr_info_entry(self):
        """
        Creates label frame, entry field and labels for DR info and sextant info with 3 columns and 4 rows

        Args:
            self: instance of PageOne class
            DR Date
            DR Time
            DR Latitude
            DR Longitude
            Pressure
            Fix Date
            Fix Time
        """
        
        # create labels
        self.dr_date_label = ttk.Label(self.dr_info_frame, text='DR Date UTC:')
        self.dr_time_label = ttk.Label(self.dr_info_frame, text='DR Time UTC:')
        self.dr_latitude_label = ttk.Label(self.dr_info_frame, text='DR Latitude:')
        self.dr_longitude_label = ttk.Label(self.dr_info_frame, text='DR Longitude:')
        self.course_label = ttk.Label(self.dr_info_frame, text='Course:')
        self.speed_label = ttk.Label(self.dr_info_frame, text='Speed kts:')
        

        # configure labels

        self.dr_date_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
        self.dr_time_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
        self.dr_latitude_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
        self.dr_longitude_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
        self.course_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
        self.speed_label.config(font=('Helvetica', 10, 'bold'), foreground='green')
                
        # create string variables
        self.dr_date = tk.StringVar(self)
        self.dr_time = tk.StringVar(self)
        self.dr_latitude = tk.StringVar(self)
        self.dr_longitude = tk.StringVar(self)
        self.course = tk.StringVar(self)
        self.speed = tk.StringVar(self)
        
        # create validation command instance 
        self.validate_number = self.register(InputChecking.validate_number)
        self.check_time_format = self.register(InputChecking.check_time_format)
        self.check_date_format = self.register(InputChecking.check_date_format)
        self.check_hs_format = self.register(InputChecking.check_hs_format)
        self.check_lat_format = self.register(InputChecking.check_lat_format)
        self.check_long_format = self.register(InputChecking.check_long_format)
        
        # create entries
        first_row_width = 12
        second_row_width = 12

        self.dr_date_entry = ttk.Entry(self.dr_info_frame, 
                                       textvariable=self.dr_date, 
                                       width=first_row_width,
                                       validate='focusout', 
                                       validatecommand=(self.check_date_format, '%P'),
                                       )
        
        self.dr_time_entry = ttk.Entry(self.dr_info_frame, 
                                       width=first_row_width,
                                       textvariable=self.dr_time, 
                                       validate='focusout',
                                       validatecommand=(self.check_time_format, '%P'))
        
        self.dr_latitude_entry = ttk.Entry(self.dr_info_frame, 
                                           width= first_row_width,
                                           textvariable=self.dr_latitude,
                                           validate='focusout',
                                           validatecommand=(self.check_lat_format, '%P'))
        
        self.dr_longitude_entry = ttk.Entry(self.dr_info_frame, 
                                            width=first_row_width,
                                            textvariable=self.dr_longitude,
                                            validate='focusout',
                                            validatecommand=(self.check_long_format, '%P'))
        
        self.course_entry = ttk.Entry(self.dr_info_frame, 
                                      textvariable=self.course, 
                                      width=second_row_width,
                                      validate='focusout',
                                      validatecommand=(self.validate_number, '%P'))
        
        self.speed_entry = ttk.Entry(self.dr_info_frame, 
                                     textvariable=self.speed, 
                                     width=second_row_width,
                                     validate='focusout',
                                     validatecommand=(self.validate_number, '%P'))



        self.dr_entry_fields = [self.dr_date_entry, 
                             self.dr_time_entry, 
                             self.dr_latitude_entry, 
                             self.dr_longitude_entry,
                             self.course_entry, 
                             self.speed_entry, 
                             ]
        
        self.dr_text_variables = [self.dr_date, 
                                  self.dr_time, 
                                  self.dr_latitude, 
                                  self.dr_longitude, 
                                  self.course, 
                                  self.speed, 
                                  ]
        
        # make the entry fields bold 
        for entry in self.dr_entry_fields:
            entry.config(font=('Helvetica', 12, 'bold'), justify='center')
        

        # Grid labels
        label_padx = 10
        label_pady = 10
        entry_padx = 10
        entry_pady = 10

        # Grid labels, put in center of label frame
        self.dr_date_label.grid(row=0, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.dr_time_label.grid(row=1, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.dr_latitude_label.grid(row=2, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.dr_longitude_label.grid(row=3, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.course_label.grid(row=4, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.speed_label.grid(row=5, column=0, sticky='E', padx=label_padx, pady=label_pady)


        # Grid entry fields
        self.dr_date_entry.grid(row=0, column=1, padx=entry_padx, pady=entry_pady)
        self.dr_time_entry.grid(row=1, column=1, padx=entry_padx, pady=entry_pady)
        self.dr_latitude_entry.grid(row=2, column=1, padx=entry_padx, pady=entry_pady)
        self.dr_longitude_entry.grid(row=3, column=1, padx=entry_padx, pady=entry_pady)
        self.course_entry.grid(row=4, column=1, padx=entry_padx, pady=entry_pady)
        self.speed_entry.grid(row=5, column=1, padx=entry_padx, pady=entry_pady)

        # Add invisible labels for alignment if needed
        self.invisble_label_dr = ttk.Label(self.dr_info_frame, text="",)
        self.invisble_label_dr.grid(row=6, column=0, padx=label_padx, pady=label_pady, sticky = 'NSEW')

        self.invisble_label_dr2 = ttk.Label(self.dr_info_frame, text="")
        self.invisble_label_dr2.grid(row=6, column=1, padx=label_padx, pady=label_pady, sticky = 'NSEW')

        self.invisble_label_dr3 = ttk.Label(self.dr_info_frame, text="")
        self.invisble_label_dr3.grid(row=6, column=2, padx=label_padx, pady=label_pady, sticky = 'NSEW')
        
      
        self.dr_info_frame.grid_columnconfigure(0, weight=1)
        self.dr_info_frame.grid_columnconfigure(1, weight=1)
        self.dr_info_frame.grid_columnconfigure(2, weight=1)
        # self.dr_info_frame.grid_columnconfigure(3, weight=1)
        # self.dr_info_frame.grid_columnconfigure(4, weight=1)
        # self.dr_info_frame.grid_columnconfigure(5, weight=1)

        # Adjust column and row weights for centering
        for i in range(6):  # Assuming 7 columns in total
            self.dr_info_frame.grid_rowconfigure(i, weight=1)
        

    def create_sextant_info_entry(self):
        
        # create labels
        self.index_error_label = ttk.Label(self.sextant_info_frame, text='Index Error :')
        self.height_of_eye_label = ttk.Label(self.sextant_info_frame, text='Height of Eye ft:')
        self.temperature_label = ttk.Label(self.sextant_info_frame, text='Temperature C:')
        self.pressure_label = ttk.Label(self.sextant_info_frame, text='Pressure mb:')
        
        # configure labels
        self.index_error_label.config(font=('Helvetica', 10, 'bold'), )
        self.height_of_eye_label.config(font=('Helvetica', 10, 'bold'), )
        self.temperature_label.config(font=('Helvetica', 10, 'bold'), )
        self.pressure_label.config(font=('Helvetica', 10, 'bold'), )

        
        # create string variables
        self.index_error = tk.StringVar(self)
        self.height_of_eye = tk.StringVar(self)
        self.temperature = tk.StringVar(self)
        self.pressure = tk.StringVar(self)

        # create entries
        self.index_error_entry = ttk.Entry(self.sextant_info_frame, 
                                           textvariable=self.index_error, 
                                           width=12,
                                           validate='focusout',
                                           validatecommand=(self.validate_number, '%P'))

        self.height_of_eye_entry = ttk.Entry(self.sextant_info_frame, 
                                             textvariable=self.height_of_eye, 
                                             width=12,
                                             validate='focusout',
                                             validatecommand=(self.validate_number, '%P'))

        self.temperature_entry = ttk.Entry(self.sextant_info_frame,
                                           width=12,
                                           textvariable=self.temperature, 
                                           validate='focusout',
                                           validatecommand=(self.validate_number, '%P'),
                                           text = '10.0'
                                           )
        

        self.pressure_entry = ttk.Entry(self.sextant_info_frame, 
                                        width=12,
                                        textvariable=self.pressure, 
                                        validate='focusout',
                                        validatecommand=(self.validate_number, '%P', ),
                                        text = '1010.0'
                                        )

        self.sextant_entry_fields = [self.index_error_entry,
                                    self.height_of_eye_entry,
                                    self.temperature_entry,
                                    self.pressure_entry] 

        # make the entry fields bold 
        for entry in self.sextant_entry_fields:
            entry.config(font=('Helvetica', 12, 'bold'), justify='center')
        
        label_padx = 10
        label_pady = 10
        entry_padx = 10
        entry_pady = 10


        # grid labels and entry fields, put in center of label frame
        self.index_error_label.grid(row=0, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.height_of_eye_label.grid(row=1, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.temperature_label.grid(row=2, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.pressure_label.grid(row=3, column=0, sticky='E', padx=label_padx, pady=label_pady)
        self.index_error_entry.grid(row=0, column=1, padx=entry_padx, pady=entry_pady)
        self.height_of_eye_entry.grid(row=1, column=1, padx=entry_padx, pady=entry_pady)
        self.temperature_entry.grid(row=2, column=1, padx=entry_padx, pady=entry_pady)
        self.pressure_entry.grid(row=3, column=1, padx=entry_padx, pady=entry_pady)

        # create invisible labels for alignment
        self.invisible_label3 = ttk.Label(self.sextant_info_frame, text="")
        self.invisible_label3.grid(row=4, column=0, padx=label_padx, pady=label_pady)

        self.invisible_label4 = ttk.Label(self.sextant_info_frame, text="")
        self.invisible_label4.grid(row=4, column=1, padx=label_padx, pady=label_pady)

        self.invisible_label5 = ttk.Label(self.sextant_info_frame, text="")
        self.invisible_label5.grid(row=4, column=2, padx=label_padx, pady=label_pady)


        # Adjust column and row weights for centering
        self.sextant_info_frame.grid_columnconfigure(0, weight=1)
        self.sextant_info_frame.grid_columnconfigure(1, weight=1)
        self.sextant_info_frame.grid_columnconfigure(2, weight=1)

        for i in range(4):  # Assuming 4 rows in total
            self.sextant_info_frame.grid_rowconfigure(i, weight=1)

    
    def create_sight_info_entry(self):
        """ 
        Creates labels and entry field for:
        Body : Autocomplete Entry
        Hs : Entry
        Date : Entry
        Time : Entry
        """
        # create labels
        self.body_label = ttk.Label(self.sight_info_entry_frame, text='Body:')
        self.hs_label = ttk.Label(self.sight_info_entry_frame, text='Hs:')
        self.date_label = ttk.Label(self.sight_info_entry_frame, text='Date:')
        self.time_label = ttk.Label(self.sight_info_entry_frame, text='Time:')

        # auto complete options
        named_bodies = ['SunLL', 'SunUL', 'MoonLL', 'MoonUL', 'Mars', 'Venus', 'Jupiter', 'Saturn']
        named_stars = [*cnav.Sight.named_star_dict]
        options = named_bodies + named_stars

        # create string variables
        self.body = tk.StringVar()
        self.hs = tk.StringVar()
        self.date = tk.StringVar()
        self.time = tk.StringVar()

        # create entry fields
        self.body_entry = AutocompleteCombobox(self.sight_info_entry_frame, completevalues=options, textvariable=self.body, width=10)
        self.hs_entry = ttk.Entry(self.sight_info_entry_frame, textvariable=self.hs, width=12)
        self.date_entry = ttk.Entry(self.sight_info_entry_frame, textvariable=self.date, width=12)
        self.time_entry = ttk.Entry(self.sight_info_entry_frame, textvariable=self.time, width=12)

        # sight entry fields
        self.sight_entry_fields = [self.body_entry, 
                                   self.hs_entry, 
                                   self.date_entry, 
                                   self.time_entry]

        # make the entry fields bold
        for entry in self.sight_entry_fields:
            entry.config(font=('Helvetica', 12, 'bold'), justify='center')
        
        # make the labels bold
        self.body_label.config(font=('Helvetica', 10, 'bold'))
        self.hs_label.config(font=('Helvetica', 10, 'bold'))
        self.date_label.config(font=('Helvetica', 10, 'bold'))
        self.time_label.config(font=('Helvetica', 10, 'bold'))

        # grid labels and entry fields, put in center of label frame
        self.body_label.grid(row=0, column=0, sticky='E', padx=10, pady=10)  # sticky changed to 'E' for right alignment
        self.hs_label.grid(row=1, column=0, sticky='E', padx=10, pady=10)
        self.date_label.grid(row=2, column=0, sticky='E', padx=10, pady=10)
        self.time_label.grid(row=3, column=0, sticky='E', padx=10, pady=10)
        self.body_entry.grid(row=0, column=1, padx=10, pady=10)  # removed sticky to allow natural width
        self.hs_entry.grid(row=1, column=1, padx=10, pady=10)
        self.date_entry.grid(row=2, column=1, padx=10, pady=10)
        self.time_entry.grid(row=3, column=1, padx=10, pady=10)

        # create add, update, and delete buttons
        self.add_button = ttk.Button(self.sight_info_entry_frame, text='Add', 
                                     command=lambda: add_new_sight(
                                            self,
                                            self.body_entry,
                                            [self.body_entry,
                                            self.hs_entry,
                                            self.date_entry,
                                            self.time_entry],
                                            self.sight_list_treeview
                                     ),
                                     style='Warning.TButton'

        )

        self.update_button = ttk.Button(self.sight_info_entry_frame, text='Update', 
                                        command=lambda: update_sight(
                                            [self.body_entry,
                                            self.hs_entry,
                                            self.date_entry,
                                            self.time_entry],
                                            self.sight_list_treeview
                                        ),
                                        style='Warning.TButton'
                                                                 
        )

        self.delete_button = ttk.Button(self.sight_info_entry_frame, text='Delete', 
                                        command = lambda: delete_sight(self.sight_list_treeview),
                                        style='Warning.TButton'
                                        
        )

         # grid buttons
        self.add_button.grid(row=4, column=0, sticky='NESW', padx=10, pady=10)
        self.update_button.grid(row=4, column=1, sticky='NESW', padx=10, pady=10)
        self.delete_button.grid(row=4, column=2, sticky='NESW', padx=10, pady=10)

        # Adjust column and row weights for centering
        self.sight_info_entry_frame.grid_columnconfigure(0, weight=1)
        self.sight_info_entry_frame.grid_columnconfigure(1, weight=1)
        self.sight_info_entry_frame.grid_columnconfigure(2, weight=1)

        for i in range(4):  # 5 rows in total
            self.sight_info_entry_frame.grid_rowconfigure(i, weight=1)


    def create_fix_info_entry(self):
        """
        Creates Treeview with fields for:
        Date, 
        Computed Lat
        Computed Long
        DR Lat
        DR Long
        """
        self.fix_info_frame = self.entry_page_4

        style = ttk.Style()
        style.configure('info.Treeview', rowheight=50) 

                # create treeview
        self.fix_treeview = ttk.Treeview(self.fix_info_frame, height=1, style = 'warning.Treeview')


        # add columns to treeview
        self.fix_treeview['columns'] = ('Date', 'Computed Lat', 'Computed Long', 'DR Lat', 'DR Long')
        self.fix_treeview.column('#0', width=0, stretch='no')
        self.fix_treeview.column('Date', anchor='center', width=220, stretch='yes')
        self.fix_treeview.column('Computed Lat', anchor='center', width=80)
        self.fix_treeview.column('Computed Long', anchor='center', width=80)
        self.fix_treeview.column('DR Lat', anchor='center', width=80)
        self.fix_treeview.column('DR Long', anchor='center', width=80)

        # add headings to treeview
        self.fix_treeview.heading('#0', text='', anchor='w')
        self.fix_treeview.heading('Date', text='Date', anchor='center')
        self.fix_treeview.heading('Computed Lat', text='Computed Lat', anchor='center')
        self.fix_treeview.heading('Computed Long', text='Computed Long', anchor='center')
        self.fix_treeview.heading('DR Lat', text='DR Lat', anchor='center')
        self.fix_treeview.heading('DR Long', text='DR Long', anchor='center')


               
        first_row_width = 10

        # create labels
        self.fix_date_label = ttk.Label(self.fix_info_frame, text='Fix Date UTC:')
        self.fix_time_label = ttk.Label(self.fix_info_frame, text='Fix Time UTC:')

        # configure labels 
        self.fix_date_label.config(font=('Helvetica', 10, 'bold'), foreground='orange')
        self.fix_time_label.config(font=('Helvetica', 10, 'bold'), foreground='orange')


        # create string variables 
        self.fix_date = tk.StringVar(self)
        self.fix_time = tk.StringVar(self)

        # create entries
        self.fix_date_entry = ttk.Entry(self.fix_info_frame,
                                        width=first_row_width,
                                        textvariable=self.fix_date,
                                        validate='focusout',
                                        validatecommand=(self.check_date_format, '%P')
                                        )

        self.fix_time_entry = ttk.Entry(self.fix_info_frame, 
                                        width=first_row_width,
                                        textvariable=self.fix_time,
                                        validate='focusout',
                                        validatecommand=(self.check_time_format, '%P')
                                        )
        
        self.fix_entry_fields = [self.fix_date_entry,
                                 self.fix_time_entry]
        
        # grid entries 
        self.fix_date_entry.grid(row=0, column=1, padx=10, pady=10)
        self.fix_time_entry.grid(row=1, column=1, padx=10, pady=10)

        # make the entry fields bold
        for entry in self.fix_entry_fields:
            entry.config(font = ('Helvetica', 12, 'bold'), justify = 'center')

        # make the labels bold
        self.fix_date_label.config(font=('Helvetica', 10, 'bold'))
        self.fix_time_label.config(font=('Helvetica', 10, 'bold'))

        # grid labels and entry fields, put in center of label frame
        self.fix_date_label.grid(row=0, column=0, sticky='E', padx=10, pady=10)
        self.fix_time_label.grid(row=1, column=0, sticky='E', padx=10, pady=10)

        
        # create ttk meter
        from ttkbootstrap.widgets import Progressbar
        self.meter = Floodgauge(self.fix_info_frame, value = 0, mode = 'determinate', maximum = 100, bootstyle = 'primary.Horizontal.TFloodgauge', mask = "Fix Confidence: {}%", )

        # grid meter
        self.meter.grid(row=2, column=0, padx=10, pady=10, columnspan=5, sticky='nsew')
        
        
        # create compute fix button
        self.compute_fix_button = ttk.Button(self.fix_info_frame, text='Compute Fix', 
                                             command=self.on_compute_fix_button_click, style = 'danger.outline.TButton')
        


        # Place Treeview in the next row, spanning across all columns
        self.fix_treeview.grid(row=3, column=0, padx=10, pady=10, columnspan=5, sticky='nsew')

        # Compute Fix Button
        self.compute_fix_button.grid(row=4, column=0, padx=10, pady=10, columnspan=5, sticky='nsew')


        # Adjust column and row weights for centering
        self.fix_info_frame.grid_columnconfigure(0, weight=1)
        self.fix_info_frame.grid_columnconfigure(1, weight=1)
        self.fix_info_frame.grid_columnconfigure(2, weight=1)

        # Adjust column and row weights for centering
        for i in range(5):  # Assuming 4 rows in total
            self.fix_info_frame.grid_rowconfigure(i, weight=1)


    def aggregate_entry_fields(self):
        self.fields = self.dr_entry_fields + self.sextant_entry_fields + self.fix_entry_fields
        
        return self.fields

    def on_compute_fix_button_click(self):

        reduction_instance = CapellaSightReduction(
        self.fields,
        [self.sight_list_treeview,
        self.fix_treeview],
        self.meter
    )
        self.master.page2.refresh_figure()
        self.master.page3.refresh_figure()


           
    def autocompletion_binding(self):
        # instantiate autocompletion class
        self.autocomplete = AutoComplete(self.master)

        # bind autocompletion to DR entry fields
        self.dr_date_entry.bind('<KeyRelease>', lambda event: self.autocomplete.date_formatting(event, self.dr_date_entry))

        self.dr_time_entry.bind('<KeyRelease>', lambda event: self.autocomplete.time_formatting(event, self.dr_time_entry))

        self.dr_latitude_entry.bind('<KeyRelease>', lambda event: self.autocomplete.lat_formatting(event, self.dr_latitude_entry))

        self.dr_longitude_entry.bind('<KeyRelease>', lambda event: self.autocomplete.long_formatting(event, self.dr_longitude_entry))

        self.hs_entry.bind('<KeyRelease>', lambda event: self.autocomplete.hs_formatting(event, self.hs_entry))

        

        # bind autocompletion to sight entry fields
        self.date_entry.bind('<KeyRelease>', lambda event: self.autocomplete.date_formatting(event, self.date_entry))

        self.time_entry.bind('<KeyRelease>', lambda event: self.autocomplete.time_formatting(event, self.time_entry))
    
    def create_tooltips(self):
        # Get the directory where the current script (presumably __main__.py or similar) is located
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # go up one directory from current_dir
        script_dir = os.path.dirname(current_dir)
        
        # Construct the absolute path to tooltips.txt
        tooltips_path = os.path.join(script_dir, 'text_files', 'tooltips.txt')

        self.extractor = TextExtractor(tooltips_path) 
        # create update button tooltip
        self.update_button_tooltip = ToolTip(self.update_button, self.extractor.get_text('updating_a_sight'))

        # create delete button tooltip
        self.delete_button_tooltip = ToolTip(self.delete_button, self.extractor.get_text('deleting_a_sight'))

        # create add button tooltip
        self.add_button_tooltip = ToolTip(self.add_button, self.extractor.get_text('adding_a_sight'))

        # create tooltip for DR info frame
        self.dr_info_frame_tooltip = ToolTip(self.dr_info_frame, self.extractor.get_text('setting_a_dr'))
        
        # create tooltop for sextant info frame
        self.sextant_info_toolop = ToolTip(self.sextant_info_frame, self.extractor.get_text('adding_sextant_info')) 
        
        # create tooltip for fix info frame
        self.fix_info_frame_tooltip = ToolTip(self.fix_info_frame, self.extractor.get_text('computing_a_fix'))
        

    
#TODO - Tooltips for Sights
#TODO - Autoformatting for dates/times
#TODO - make Sight planning controls larger 
#TODO - make sight planning controls more intuitive
#TODO - fix multiple message box screen shift bug
#TODO - Add averaging tooltip 


class AutoComplete:
    '''Formats user input to match the format of the database'''

    def __init__(self, master):
        self.master = master

    def _insert_at_positions(self, entry, char, positions):
        '''Insert character at specific positions in the entry'''
        for pos in positions:
            if len(entry.get()) == pos:
                entry.insert(pos+1, char)

    def time_formatting(self, event, entry):
        '''Formats time input as hh:mm:ss'''
        self._insert_at_positions(entry, ':', [2, 5])

    def date_formatting(self, event, entry):
        '''Formats date input as yyyy-mm-dd'''
        self._insert_at_positions(entry, '-', [4, 7])

    def hs_formatting(self, event, entry):
        '''Formats hs input as dd-mm.t'''
        self._insert_at_positions(entry, '-', [2])
        self._insert_at_positions(entry, '.', [5])

    def lat_formatting(self, event, entry):
        '''Formats latitude input as dd-mm.t-N/S'''
        self._insert_at_positions(entry, '-', [2])
        self._insert_at_positions(entry, '.', [5])
        self._insert_at_positions(entry, '-', [7])
        # autocorrects lower cases
        lat_val = entry.get()
        if len(lat_val) == 9 and lat_val[-1] in ['n', 's']:
            entry.delete(8)
            entry.insert(9, lat_val[-1].upper())

    def long_formatting(self, event, entry):
        '''Formats longitude input as ddd-mm.t-E/W'''
        self._insert_at_positions(entry, '-', [3, 8])
        self._insert_at_positions(entry, '.', [6])
        # autocorrects lower cases
        long_val = entry.get()
        if len(long_val) == 10 and long_val[-1] in ['e', 'w']:
            entry.delete(9)
            entry.insert(10, long_val[-1].upper())



import numpy as np


class DRCalc():
    """
    dr_calc() : will calculate a mercator sailing based on a position,
    C/S and a dt.dimedelta object. It can DR from a position either forwards or backwards
    using the recriprocal of the DR course. This is useful for the sight analysis functions,
    since the function starts from the Least Squares fix and then DR's backwards to the time of each sight,
    this is allows the initial DR position to be extremely inaccurate, yet
    still provide an effective fit-slope analysis.
    """

    def __init__(self, init_lat, init_long, timedelta, course, speed):
        self.init_lat = float(init_lat)
        self.init_long = float(init_long)
        self.timedelta = float(timedelta) / 3600
        self.course = float(course)
        self.speed = float(speed)
        self.dr_coord_calc_fwd()
        self.dr_coord_calc_bwd()

        return

    def dr_coord_calc_fwd(self):
        self.distance = self.timedelta * self.speed
        if self.course == 90:
            self.lat2 = self.init_lat
            self.dlo = (self.distance / np.cos(np.deg2rad(self.init_lat))) / 60
        elif self.course == 270:
            self.lat2 = self.init_lat
            self.dlo = -1 * (self.distance / np.cos(np.deg2rad(self.init_lat))) / 60
        else:
            if 0 < self.course < 90:
                self.courseangle = self.course
            elif 90 < self.course < 180:
                self.courseangle = 180 - self.course
            elif 180 < self.course < 270:
                self.courseangle = self.course + 180
            else:
                self.courseangle = 360 - self.course
            self.lat2 = (self.distance * np.cos(np.deg2rad(self.course))) / 60 + self.init_lat
            mpartsinitial = 7915.7045 * np.log10(
                np.tan(np.pi / 4 + (np.deg2rad(self.init_lat) / 2))) - 23.2689 * np.sin(
                np.deg2rad(self.init_lat))
            mpartssecond = 7915.7045 * np.log10(np.tan(np.pi / 4 + (np.deg2rad(self.lat2) / 2))) - 23.2689 * np.sin(
                np.deg2rad(self.lat2))
            littlel = mpartssecond - mpartsinitial
            self.dlo = (littlel * np.tan(np.deg2rad(self.course))) / 60
        self.drlatfwds = self.lat2
        self.drlongfwds = self.init_long + self.dlo
        if self.drlongfwds >= 180:
            self.drlongfwds = self.drlongfwds - 360

        return

    def dr_coord_calc_bwd(self):

        self.course = (self.course - 180) % 360
        self.distance = self.timedelta * self.speed
        if self.course == 90:
            self.lat2 = self.init_lat
            self.dlo = (self.distance / np.cos(np.deg2rad(self.init_lat))) / 60
        elif self.course == 270:
            self.lat2 = self.init_lat
            self.dlo = -1 * (self.distance / np.cos(np.deg2rad(self.init_lat))) / 60
        else:
            if 0 < self.course < 90:
                self.courseangle = self.course
            elif 90 < self.course < 180:
                self.courseangle = 180 - self.course
            elif 180 < self.course < 270:
                self.courseangle = self.course + 180
            else:
                self.courseangle = 360 - self.course

            self.lat2 = (self.distance * np.cos(np.deg2rad(self.course))) / 60 + self.init_lat
            mpartsinitial = 7915.7045 * np.log10(
                np.tan(np.pi / 4 + (np.deg2rad(self.init_lat) / 2))) - 23.2689 * np.sin(
                np.deg2rad(self.init_lat))
            mpartssecond = 7915.7045 * np.log10(np.tan(np.pi / 4 + (np.deg2rad(self.lat2) / 2))) - 23.2689 * np.sin(
                np.deg2rad(self.lat2))
            littlel = mpartssecond - mpartsinitial
            self.dlo = (littlel * np.tan(np.deg2rad(self.course))) / 60

        self.drlatbackwards = self.lat2
        self.drlongbackwards = self.init_long + self.dlo
        if self.drlongbackwards >= 180:
            self.drlongbackwards = self.drlongbackwards - 360
        return




import tkinter as tk
import ttkbootstrap as ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import utilities.celestial_engine as celestial_engine

class FitSlopePage(ttk.Frame):
    def __init__(self, container):
        super().__init__(container)

        self.fit_slope_plot()
        self.draw_canvas()
        

    def fit_slope_plot(self):
        #create canvas
        self.canvas_fit_slope = tk.Canvas(self, width=800, height=800)

        # get figure from lop_plot 
        self.fit_slope_scatter_plot = celestial_engine.plt.figure(1)
        self.fit_slope_scatter_plot.set_facecolor('#222222')

        # add figure to canvas
        self.fit_slope_scatter_plot_canvas = FigureCanvasTkAgg(self.fit_slope_scatter_plot, master=self.canvas_fit_slope)
        
        # add toolbar to canvas
        self.fit_slope_scatter_plot_toolbar = NavigationToolbar2Tk(self.fit_slope_scatter_plot_canvas, self.canvas_fit_slope)

        # update
        self.fit_slope_scatter_plot_toolbar.update()

        # packing order is important! 
        self.fit_slope_scatter_plot_toolbar.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        
        self.canvas_fit_slope.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        self.fit_slope_scatter_plot_canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
   
    def draw_canvas(self):
        # draw canvas
        self.fit_slope_scatter_plot_canvas.draw()

    def refresh_figure(self):
        self.draw_canvas()
